@@ -13,12 +13,12 @@ warn <- function(f, wmsg="", ...) {
 
 messages <- function(x, f="") {
 	#g <- gc(verbose=FALSE)
-	if (methods::.hasSlot(x, "ptr")) {
-		if (x@ptr$has_warning()) {
-			warn(f, paste(unique(x@ptr$getWarnings()), collapse="\n"))
+	if (methods::.hasSlot(x, "cpp")) {
+		if (x@cpp$has_warning()) {
+			warn(f, paste(unique(x@cpp$getWarnings()), collapse="\n"))
 		}
-		if (x@ptr$has_error()) {
-			error(f, x@ptr$getError())
+		if (x@cpp$has_error()) {
+			error(f, x@cpp$getError())
 		}
 	} else {
 		if (x$has_warning()) {
@@ -37,7 +37,7 @@ mem_info <- function(x, n=1) {
 	n <- max(0,n)
 	opt <- spatOptions()
 	opt$ncopies = n;
-	v <- x@ptr$mem_needs(opt)
+	v <- x@cpp$mem_needs(opt)
 	memmin <- opt$memmin
 	memmax <- opt$memmax
 	#if (print) {
@@ -46,7 +46,7 @@ mem_info <- function(x, n=1) {
 		cat("\nMemory (GB) ")
 		cat("\n------------------------")
 
-		cat(paste("\ncheck threshold :", opt$memmin / gb, " (memmin)"))
+		cat(paste("\ncheck threshold :", opt$memmin / gb, "(memmin)"))
 		if (memmax > 0) {
 			cat(paste("\navailable       :",  round(v[2] / gb, 2), "(memmax)"))
 		} else {
@@ -68,7 +68,7 @@ mem_info <- function(x, n=1) {
 free_RAM <- function() {
 	opt <- spatOptions()
 	x <- rast()
-	v <- x@ptr$mem_needs(opt)
+	v <- x@cpp$mem_needs(opt)
 	v[2] / 128
 }
 
