@@ -54,11 +54,7 @@
 
 
 
-.write_cdf <- function(x, filename, overwrite=FALSE, zname="time", atts="", gridmap="", prec="float", compression=NA, missval, ...) {
-
-	dots <- list(...)
-	force_v4 <- isTRUE(dots$force_v4)
-	verbose  <- isTRUE(dots$verbose)
+.write_cdf <- function(x, filename, overwrite=FALSE, zname="time", atts="", gridmap="", prec="float", compression=NA, missval, force_v4=TRUE, verbose=FALSE, ...) {
 
 	n <- length(x)
 	y <- x[1]
@@ -101,23 +97,28 @@
 	ncvars <- list()
 	cal <- NA
 	for (i in 1:n) {
-		if ((nl[i] > 1) || (x[i]@ptr$hasTime)) {
+		if ((nl[i] > 1) || (x[i]@cpp$hasTime)) {
 			y <- x[i]
-			if (y@ptr$hasTime) {
-				zv <- y@ptr$time
-				tstep <- y@ptr$timestep
+			if (y@cpp$hasTime) {
+				zv <- y@cpp$time
+				tstep <- y@cpp$timestep
 				cal <- "standard"
 				if (tstep == "seconds") {
 					zunit <- "seconds since 1970-1-1 00:00:00"
-					cal <- "standard"
 				} else if (tstep == "days") {
 					zunit <- "days since 1970-1-1"
 					zv <- zv / (24 * 3600)
-					cal <- "standard"
+				} else if (tstep == "months") {
+					zunit <- "months"
+					zv <- time(y)
+				} else if (tstep == "yearmonths") {
+					zunit <- "months since 1970"
+					tm <- time(y) - 1970
+					yr <- tm %/% 1
+					zv <- (yr*12) + round(12 * (tm %% 1))
 				} else if (tstep == "years") {
 					zunit <- "years since 1970"
 					zv <- time(y) - 1970
-					cal <- "standard"
 				} else {
 					zunit <- "unknown"
 				}
@@ -133,7 +134,6 @@
 		}
 	}
 
-
 	ncvars[[n+1]] <- ncdf4::ncvar_def("crs", "", list(), NULL, prec="integer")
 	
 	ncobj <- ncdf4::nc_create(filename, ncvars, force_v4=force_v4, verbose=verbose)
@@ -215,17 +215,39 @@
 
 
 setMethod("writeCDF", signature(x="SpatRaster"),
-	function(x, filename, varname, longname="", unit="", ...) {
+	function(x, filename, varname, longname="", unit="", split=FALSE, ...) {
 		filename <- trimws(filename)
+		if (length(filename) > 1) {
+			if (length(filename) != nlyr(x)) {
+				stop("either provide a single filename, or the same number as nlyr(x)")
+			}
+		}
 		stopifnot(filename != "")
-		if (missing(varname)) {
-			varname <- tools::file_path_sans_ext(basename(filename))
+		if (split) {
+			y <- sds(as.list(x))
+			names(y) <- names(x)
+			if (!missing(varname)) {
+				varnames(y) <- varname
+			}
+			if (!missing(longname)) {
+				longnames(y) <- longname
+			}
+			if (!missing(unit)) {
+				units(y) <- unit
+			} else {
+				units(y) <- units(x)
+			}
+			invisible( writeCDF(y, filename=filename, ...) )
+		} else {
+			if (missing(varname)) {
+				varname <- tools::file_path_sans_ext(basename(filename))
+			}
+			varnames(x) <- varname
+			longnames(x) <- longname
+			units(x) <- unit
+			x <- sds(x)
+			invisible( writeCDF(x, filename=filename, ...) )
 		}
-		varnames(x) <- varname
-		longnames(x) <- longname
-		units(x) <- unit
-		x <- sds(x)
-		invisible( writeCDF(x, filename=filename, ...) )
 	}
 )
 
