@@ -1,17 +1,23 @@
-
 .terra_environment <- new.env(parent=emptyenv())
 
 
 .create_options <- function() {
 	opt <- methods::new("SpatOptions")
-	opt@ptr <- SpatOptions$new()
+	opt@cpp <- SpatOptions$new()
 	# check=T does not exist in ancient R
 	tmpdir <- try(tempdir(check = TRUE), silent=TRUE)
-	opt@ptr$tempdir <- normalizePath(tempdir(), winslash="/")
+	opt@cpp$tempdir <- normalizePath(tempdir(), winslash="/")
 	.terra_environment$options <- opt
+	.terra_environment$devs <- NULL
+
+	x <- options("terra_default")[[1]]
+	if (!is.null(x)) {
+		do.call(terraOptions, x)
+	}
+
 }
 
-.options_names <- function() {
+.option_names <- function() {
 	c("progress", "progressbar", "tempdir", "memfrac", "memmax", "memmin", "datatype", "filetype", "filenames", "overwrite", "todisk", "names", "verbose", "NAflag", "statistics", "steps", "ncopies", "tolerance", "pid", "threads", "scale", "offset") #, "append")
 }
 
@@ -30,7 +36,7 @@
 		x$gdal_options <- gopt
 	}
 
-	s <- nms %in% .options_names()
+	s <- nms %in% .option_names()
 
 	if (any(!s)) {
 		bad <- paste(nms[!s], collapse=",")
@@ -72,7 +78,7 @@
 defaultOptions <- function() {
 	## work around onLoad problem
 	if (is.null(.terra_environment$options)) .create_options()
-	.terra_environment$options@ptr$deepcopy()
+	.terra_environment$options@cpp$deepcopy()
 }
 
 spatOptions <- function(filename="", overwrite=FALSE, ..., wopt=NULL) {
@@ -82,9 +88,9 @@ spatOptions <- function(filename="", overwrite=FALSE, ..., wopt=NULL) {
 	## work around onLoad problem
 	if (is.null(.terra_environment$options)) .create_options()
 
-	opt <- .terra_environment$options@ptr$deepcopy()
+	opt <- .terra_environment$options@cpp$deepcopy()
 	opt$pid <- Sys.getpid()
-	filename <- .fullFilename(filename, mustExist=FALSE)
+	filename <- .fullFilename(filename, mustExist=TRUE)
 	if (!is.null(unlist(wopt))) {
 		wopt$filenames <- filename
 		wopt$overwrite <- overwrite[1]
@@ -121,7 +127,8 @@ spatOptions <- function(filename="", overwrite=FALSE, ..., wopt=NULL) {
 	nms <- names(opt)
 	nms <- nms[!grepl("^\\.", nms)]
 	nms <- nms[!(nms %in% c("initialize", "messages", "getClass", "finalize", "datatype_set", "pid", "statistics", "gdal_options", "scale", "offset", "threads", "filenames", "NAflag"))]
-	nms <- gsub("def_", "", nms)
+	defnms <- grepl("^def_", nms)
+	nms <- nms[!defnms]
 	out <- sapply(nms, function(n) eval(parse(text=paste0("opt$", n))))
 	out$memmin <- 8 * out$memmin / (1024^3)
 	if (out$memmax > 0) {
@@ -148,41 +155,53 @@ spatOptions <- function(filename="", overwrite=FALSE, ..., wopt=NULL) {
 	c("datatype", "filetype") #, "verbose")
 }
 
+
 terraOptions <- function(..., print=TRUE) {
 	dots <- list(...)
 	if (is.null(.terra_environment$options)) .create_options()
-	opt <- .terra_environment$options@ptr
+	opt <- .terra_environment$options@cpp
 
 	nms <- names(dots)
-	ndots <- length(dots)
+
+	if (length(dots) == 0) {
+		return(.showOptions(opt, print=print))
+	}
+
+	ok <- nms %in% .option_names()
+	if (any(!ok)) {
+		bad <- paste(nms[!ok], collapse=", ")
+		warn("terraOptions<-", paste("unknown option(s):", bad))
+		dots <- dots[ok]
+		nms <- nms[ok]			
+		if (length(dots) == 0) return(invisible())
+	}
 
 	if ("tempdir" %in% nms) {
 		i <- which(nms == "tempdir")
+		dots[i] <- path.expand(trimws(dots[i]))
 		if (!dir.exists(dots[[i]])) {
 			warn("options", "you cannot set the tempdir to a path that does not exist")
 			dots <- dots[-i]
 			nms <- nms[-i]
+			if (length(dots) == 0) return(invisible())
 		}
 	}
 
-	if (ndots == 0) {
-		.showOptions(opt, print=print)
-	} else if (length(dots) > 0) {
-		d <- nms %in% .default_option_names()
-		dnms <- paste0("def_", nms)
-		for (i in 1:length(nms)) {
-			if (d[i]) {
-				opt[[ dnms[i] ]] <- dots[[ i ]]
-			} else {
-				opt[[ nms[i] ]] <- dots[[ i ]]
-			}
+	d <- nms %in% .default_option_names()
+	dnms <- paste0("def_", nms)
+	for (i in 1:length(nms)) {
+		if (d[i]) {
+			opt[[ dnms[i] ]] <- dots[[ i ]]
+		} else {
+			opt[[ nms[i] ]] <- dots[[ i ]]
 		}
-		if ("memfrac" %in% nms) {
-			if (dots$memfrac > 0.9) {
-				warn("terraOptions", "memfrac > 0.9")
-			}
-		}
-		.terra_environment$options@ptr <- opt
 	}
+		
+	if ("memfrac" %in% nms) {
+		if (dots$memfrac > 0.9) {
+			warn("terraOptions", "memfrac > 0.9")
+		}
+	}
+	.terra_environment$options@cpp <- opt
 }
 
