@@ -32,7 +32,7 @@ new_rast <- function(nrows=10, ncols=10, nlyrs=1, xmin=0, xmax=1, ymin=0, ymax=1
 	#check_proj4_datum(crs)
 
 	r <- methods::new("SpatRaster")
-	r@ptr <- SpatRaster$new(c(nrows, ncols, nlyrs), e, crs)
+	r@cpp <- SpatRaster$new(c(nrows, ncols, nlyrs), e, crs)
 	r <- messages(r, "rast")
 
 	if (!missing(resolution)) {
@@ -84,7 +84,7 @@ setMethod("rast", signature(x="list"),
 		}
 		opt <- spatOptions()
 		for (i in 2:length(x)) {
-			out@ptr$addSource(x[[i]]@ptr, warn, opt)
+			out@cpp$addSource(x[[i]]@cpp, warn, opt)
 		}
 		out <- messages(out, "rast")
 		lnms <- names(x)
@@ -135,16 +135,28 @@ setMethod("rast", signature(x="SpatVector"),
 
 
 
-.fullFilename <- function(x, mustExist=TRUE) {
+.fullFilename <- function(x, mustExist=FALSE, vsi=FALSE) {
 	x <- trimws(x)
-	p <- normalizePath(x, winslash = "/", mustWork = FALSE)
+	x <- x[x != ""]
+	x <- enc2utf8(x)
+	
+	i <- substr(x, 1, 5) == "s3://" 
+	x[i] <- paste0("/vsis3/", substr(x[i], 6, nchar(x[i])))
+	if (all(i)) return(x)
+	
+	i <- substr(x, 1, 4) == "http" 
+	if (vsi) {
+		x[i] <- paste0("/vsicurl/", x[i])
+	}
+	if (all(i)) return(x)
+
+	i <- grepl(":", x)
+	if (all(i)) return(x)
+	
+	p <- normalizePath(x[!i], winslash = "/", mustWork = FALSE)
 	if (mustExist) {
-		i <- file.exists(p)
-		if (all(i)) {
-			return(p)
-		} else {
-			x[i] <- p[i]
-		}
+		j <- file.exists(dirname(p))
+		x[j] <- p[j]
 	} else {
 		return(p)
 	}
@@ -158,37 +170,35 @@ setMethod("rast", signature(x="SpatVector"),
 }
 
 setMethod("rast", signature(x="character"),
-	function(x, subds=0, lyrs=NULL, drivers=NULL, opts=NULL, win=NULL) {
+	function(x, subds=0, lyrs=NULL, drivers=NULL, opts=NULL, win=NULL, snap="near", vsi=FALSE, raw=FALSE) {
 
-		x <- trimws(x)
-		x <- x[x!=""]
-		if (length(x) == 0) {
+		f <- .fullFilename(x, vsi=vsi)
+		if (length(f) == 0) {
 			error("rast", "filename is empty. Provide a valid filename")
 		}
-		f <- .fullFilename(x)
-		f <- enc2utf8(f)
 
-#		if (tolower(tools::file_ext(f)) == "rds") {
-#			r <- readRDS(x)
-#			if (!inherits(r, "SpatRaster")) {
-#				error("rast", "the rds file does not store a SpatRaster")
-#			}
-#			return(r)
-#		}
+		if ((length(f) == 1) && grepl("\\.rds$", tolower(f[1]))) {
+			r <- unwrap(readRDS(x))
+			if (!inherits(r, "SpatRaster")) {
+				error("rast", "the rds file does not store a SpatRaster")
+			}
+			return(r)
+		}
 		
 		r <- methods::new("SpatRaster")
 		#subds <- subds[1]
 		if (is.null(opts)) opts <- ""[0]
+		if (raw) opts <- c(opts, "so=false")
 		if (is.null(drivers)) drivers <- ""[0]
 		if (length(subds) == 0) subds = 0
 		if (is.character(subds)) {
-			#r@ptr <- SpatRaster$new(f, -1, subds, FALSE, 0[])
-			r@ptr <- SpatRaster$new(f, -1, subds, FALSE, drivers, opts, 0[])
+			#r@cpp <- SpatRaster$new(f, -1, subds, FALSE, 0[])
+			r@cpp <- SpatRaster$new(f, -1, subds, FALSE, drivers, opts, 0[])
 		} else {
-			r@ptr <- SpatRaster$new(f, subds-1, "", FALSE, drivers, opts, 0[])
+			r@cpp <- SpatRaster$new(f, subds-1, "", FALSE, drivers, opts, 0[])
 		}
 		r <- messages(r, "rast")
-		if (r@ptr$getMessage() == "ncdf extent") {
+		if (r@cpp$getMessage() == "ncdf extent") {
 			# could have used opts="IGNORE_XY_AXIS_NAME_CHECKS=YES"
 			test <- try(r <- .ncdf_extent(r, f), silent=TRUE)
 			if (inherits(test, "try-error")) {
@@ -208,7 +218,9 @@ setMethod("rast", signature(x="character"),
 			r <- r[[lyrs]]
 		} 
 		if (!is.null(win)) {
-			window(r) <- win
+			e <- ext(win)
+			e <- align(e, r, snap=snap)
+			window(r) <- e
 		}
 		r
 	}
@@ -230,11 +242,11 @@ multi <- function(x, subds=0, xyz=3:1, drivers=NULL, opts=NULL) {
 	subds <- subds[1]
 
 	if (is.character(subds)) {
-		r@ptr <- SpatRaster$new(f, -1, subds, TRUE, drivers, opts, xyz-1)
+		r@cpp <- SpatRaster$new(f, -1, subds, TRUE, drivers, opts, xyz-1)
 	} else {
-		r@ptr <- SpatRaster$new(f, subds-1, ""[0], TRUE, drivers, opts, xyz-1)
+		r@cpp <- SpatRaster$new(f, subds-1, ""[0], TRUE, drivers, opts, xyz-1)
 	}
-	if (r@ptr$getMessage() == "ncdf extent") {
+	if (r@cpp$getMessage() == "ncdf extent") {
 		test <- try(r <- .ncdf_extent(r), silent=TRUE)
 		if (inherits(test, "try-error")) {
 			warn("rast", "GDAL did not find an extent. Cells not equally spaced?")
@@ -252,11 +264,11 @@ multi <- function(x, subds=0, xyz=3:1, drivers=NULL, opts=NULL) {
 
 
 setMethod("rast", signature(x="SpatRaster"),
-	function(x, nlyrs=nlyr(x), names, vals, keeptime=TRUE, keepunits=FALSE, props=FALSE) {
+	function(x, nlyrs=nlyr(x), names, vals, keeptime=TRUE, keepunits=FALSE, props=FALSE, tags=FALSE) {
 		if (inherits(nlyrs, "SpatRaster")) {
 			error("rast", "use 'c()' to combine SpatRasters")
 		}
-		x@ptr <- x@ptr$geometry(nlyrs, props, keeptime, keepunits)
+		x@cpp <- x@cpp$geometry(nlyrs, props, keeptime, keepunits, tags)
 		x <- messages(x, "rast")
 		if (!missing(names)) {
 			if (length(names) == nlyr(x)) names(x) <- names
@@ -277,7 +289,7 @@ setMethod("rast", signature(x="SpatRasterDataset"),
 			x[1]
 		} else {
 			r <- methods::new("SpatRaster")
-			r@ptr <- x@ptr$collapse()
+			r@cpp <- x@cpp$collapse()
 			nms <- names(x)
 			if (any(nms != "")) {
 				names(r) <- paste(rep(nms, nlyr(x)), names(r), sep="_")
@@ -291,6 +303,9 @@ setMethod("rast", signature(x="SpatRasterDataset"),
 setMethod("rast", signature(x="array"),
 	function(x, crs="", extent=NULL) {
 		dims <- dim(x)
+		if (length(dims) < 3) {
+			error("rast,array", "cannot handle an array with less than 3 dimensions")
+		}
 		if (length(dims) > 3) {
 			if (length(dims) == 4) {
 				if (dims[4] == 1) {
@@ -309,7 +324,7 @@ setMethod("rast", signature(x="array"),
 			e <- c(0, dims[2], 0, dims[1])
 		}
 		crs <- character_crs(crs, "rast")
-		r@ptr <- SpatRaster$new(dims, e, crs)
+		r@cpp <- SpatRaster$new(dims, e, crs)
 		values(r) <- x
 		messages(r, "rast")
 	}
@@ -405,20 +420,51 @@ setMethod("rast", signature(x="ANY"),
 	}
 
 	if (!is.null(extent)) {
-		ext(r) <- extent
+		r <- extend(r, extent)
+		r <- crop(r, extent)
 	}
 
 	return(r)
 }
 
 
+.rastFromXYLZ <- function(x, digits=6, crs="", extent=NULL) {
+	if (ncol(x) != 4) {
+		error("rast", "a 'xylz' structure must have 4 columns")
+	}
+	nms <- names(x)
+	names(x)[1:3] <- c("x", "y", "l")
+	w <- stats::reshape(x, timevar="l", idvar=c("x", "y"), direction="wide")
+	w <- rast(w, type="xyz", digits=digits, crs=crs, extent=extent)
+	names(w) <- gsub(paste0(nms[4], "."), "", names(w))
+	if (inherits(x[,3], "Date") || inherits(x[,3], "POSIXlt") || inherits(x[,3], "POSIXct")) {
+		time(w) <- unique(x[,3])
+		names(w) <- paste0(nms[4], ".", 1:nlyr(w))
+	} else if (inherits(x[,3], "numeric")) {
+		u <- unique(x[,3])
+		if (all(u == trunc(u))) {
+			su <- sort(u)
+			if ((su[1] == 1) && (su[length(su)] == length(su))) {
+				if (!all(su == u)) {
+					w <- w[[order(u)]]
+				}
+			}
+		}
+	}
+	w
+}
 
+	
 
 setMethod("rast", signature(x="matrix"),
 	function(x, type="", crs="", digits=6, extent=NULL) {
 		stopifnot(prod(dim(x)) > 0)
 		if (type == "xyz") {
 			r <- .rastFromXYZ(x, crs=crs, digits=digits, extent=extent)
+		} else if (type == "xylz") {
+			r <- .rastFromXYLZ(x, crs=crs, digits=digits, extent=extent)
+		} else if (type != "") {
+			error("rast", 'argument type should be one of "", "xyz", or "xylz"')
 		} else {
 			if (is.null(extent)) {
 				r <- rast(nrows=nrow(x), ncols=ncol(x), extent=ext(c(0, ncol(x), 0, nrow(x))), crs=crs)
@@ -436,6 +482,10 @@ setMethod("rast", signature(x="data.frame"),
 	function(x, type="xyz", crs="", digits=6, extent=NULL) {
 		if (type == "xyz") {
 			.rastFromXYZ(x, crs=crs, digits=digits, extent=extent)
+		} else if (type == "xylz") {
+			r <- .rastFromXYLZ(x, crs=crs, digits=digits, extent=extent)
+		} else if (type != "") {
+			error("rast", 'argument type should be one of "", "xyz", or "xylz"')
 		} else {
 			rast(as.matrix(x), type=type, crs=crs, digits=digits, extent=extent)
 		}
@@ -468,7 +518,7 @@ setMethod("rast", signature(x="stars_proxy"),
 setMethod("NAflag<-", signature(x="SpatRaster"),
 	function(x, value)  {
 		value <- as.numeric(value)
-		if (!(x@ptr$setNAflag(value))) {
+		if (!(x@cpp$setNAflag(value))) {
 			error("NAflag<-", "cannot set this value")
 		}
 		x
@@ -477,7 +527,7 @@ setMethod("NAflag<-", signature(x="SpatRaster"),
 
 setMethod("NAflag", signature(x="SpatRaster"),
 	function(x)  {
-		x@ptr$getNAflag()
+		x@cpp$getNAflag()
 	}
 )
 
