@@ -5,7 +5,7 @@
 
 setMethod("hasValues", signature(x="SpatRaster"),
 	function(x) {
-		x@ptr$hasValues
+		x@cpp$hasValues
 	}
 )
 
@@ -44,8 +44,8 @@ setMethod("hasValues", signature(x="SpatRaster"),
 	}
 	dd <- !(bb | ii | ff)
 	if (any(dd)) {
-		d = which(dd)
-		v[,d] = replace(v[,d], is.na(v[,d]), NA)
+		d <- which(dd)
+		v[,d] <- replace(v[,d], is.na(v[,d]), NA)
 	}
 	v
 }
@@ -55,14 +55,25 @@ setMethod("readValues", signature(x="SpatRaster"),
 function(x, row=1, nrows=nrow(x), col=1, ncols=ncol(x), mat=FALSE, dataframe=FALSE, ...) {
 	stopifnot(row > 0 && nrows > 0)
 	stopifnot(col > 0 && ncols > 0)
-	v <- x@ptr$readValues(row-1, nrows, col-1, ncols)
+	v <- x@cpp$readValues(row-1, nrows, col-1, ncols)
 	messages(x, "readValues")
-	if (dataframe || mat) {
+	if (dataframe) {
 		v <- matrix(v, ncol = nlyr(x))
 		colnames(v) <- names(x)
-		if (dataframe) {
-			return(.makeDataFrame(x, v, ...) )
+		return(.makeDataFrame(x, v, ...) )
+	} else if (mat) {
+		if (all(is.int(x))) {
+			v <- matrix(as.integer(v), ncol = nlyr(x))
+		} else if (all(is.bool(x))) {
+			v <- matrix(as.logical(v), ncol = nlyr(x))		
+		} else {
+			v <- matrix(v, ncol = nlyr(x))		
 		}
+		colnames(v) <- names(x)
+	} else if (all(is.int(x))) {
+		v <- as.integer(v)
+	} else if (all(is.bool(x))) {
+		v <- as.logical(v)
 	}
 	v
 }
@@ -104,7 +115,7 @@ setMethod("focalValues", signature("SpatRaster"),
 		readStart(x)
 		on.exit(readStop(x))
 		opt <- spatOptions()
-		m <- matrix(x@ptr$focalValues(w, fill, max(0, row-1), nrows, opt), ncol=prod(w), byrow=TRUE)
+		m <- matrix(x@cpp$focalValues(w, fill, max(0, row-1), nrows, opt), ncol=prod(w), byrow=TRUE)
 		messages(x, "focalValues")
 		m
 	}
@@ -160,7 +171,7 @@ setMethod("setValues", signature("SpatRaster"),
 		make_factor <- FALSE
 		set_coltab <- FALSE
 		if (is.character(values)) {
-			if (all(substr(na.omit(values), 1, 1) == "#")) {
+			if (all(substr(stats::na.omit(values), 1, 1) == "#")) {
 				fv <- as.factor(values)
 				if (length(levels(fv)) <= 256) {
 					values <- as.integer(fv) #-1
@@ -194,7 +205,7 @@ setMethod("setValues", signature("SpatRaster"),
 		opt <- spatOptions()
 
 		if (lv == 1) {
-			y@ptr$setValues(values, opt)
+			y@cpp$setValues(values, opt)
 		} else {
 			if (lv > (nc * nl)) {
 				warn("setValues", "values is larger than the size of cells")
@@ -203,7 +214,7 @@ setMethod("setValues", signature("SpatRaster"),
 				warn("setValues", "values were recycled")
 				values <- rep(values, length.out=nc*nl)
 			}
-			y@ptr$setValues(values, opt)
+			y@cpp$setValues(values, opt)
 		}
 		y <- messages(y, "setValues")
 		if (make_factor) {
@@ -216,9 +227,9 @@ setMethod("setValues", signature("SpatRaster"),
 		if (set_coltab) {
 			coltab(y) <- fv
 		} else if (is.logical(values)) {
-			if (!all(is.na(values))) y@ptr$setValueType(3)
+			if (!all(is.na(values))) y@cpp$setValueType(3)
 		} else if (is.integer(values)) {
-			y@ptr$setValueType(1)
+			y@cpp$setValueType(1)
 		}
 		y
 	}
@@ -228,7 +239,7 @@ setMethod("setValues", signature("SpatRaster"),
 
 setMethod("inMemory", signature(x="SpatRaster"),
 	function(x, bylayer=FALSE) {
-		r <- x@ptr$inMemory
+		r <- x@cpp$inMemory
 		if (bylayer) {
 			nl <- .nlyrBySource(x)
 			r <- rep(r, nl)
@@ -238,26 +249,30 @@ setMethod("inMemory", signature(x="SpatRaster"),
 )
 
 
-#..hasValues <- function(x) { x@ptr$hasValues}
-#..inMemory <- function(x) { x@ptr$inMemory }
-#..filenames <- function(x) {	x@ptr$filenames }
+#..hasValues <- function(x) { x@cpp$hasValues}
+#..inMemory <- function(x) { x@cpp$inMemory }
+#..filenames <- function(x) {	x@cpp$filenames }
 
+subsetSource <- function(x, i) {
+	x@cpp <- x@cpp$subsetSource(i-1)
+	messages(x)
+}
 
 setMethod("sources", signature(x="SpatRaster"),
 	function(x, nlyr=FALSE, bands=FALSE) {
-		src <- x@ptr$filenames()
+		src <- x@cpp$filenames()
 		Encoding(src) <- "UTF-8"
 		if (bands) {
-			nls <- x@ptr$nlyrBySource()
+			nls <- x@cpp$nlyrBySource()
 			d <- data.frame(sid=rep(1:length(src), nls),
 						 source=rep(src, nls),
-						 bands=x@ptr$getBands()+1, stringsAsFactors=FALSE)
+						 bands=x@cpp$getBands()+1, stringsAsFactors=FALSE)
 			if (nlyr) {
 				d$nlyr <- rep(nls, nls)
 			}
 			d
 		} else if (nlyr) {
-			data.frame(source=src, nlyr=x@ptr$nlyrBySource(), stringsAsFactors=FALSE)
+			data.frame(source=src, nlyr=x@cpp$nlyrBySource(), stringsAsFactors=FALSE)
 		} else {
 			src
 		}
@@ -283,7 +298,7 @@ setMethod("sources", signature(x="SpatRasterDataset"),
 			x <- lapply(1:length(x), function(i) cbind(cid=i, x[[i]]))
 			do.call(rbind, x)
 		} else {
-			x@ptr$filenames()
+			x@cpp$filenames()
 		}
 	}
 )
@@ -291,11 +306,11 @@ setMethod("sources", signature(x="SpatRasterDataset"),
 
 setMethod("sources", signature(x="SpatVector"),
 	function(x) {
-		if (x@ptr$source != "") {
-			if (x@ptr$layer != tools::file_path_sans_ext(basename(x@ptr$source))) {
-				paste0(x@ptr$source, "::", x@ptr$layer)
+		if (x@cpp$source != "") {
+			if (x@cpp$layer != tools::file_path_sans_ext(basename(x@cpp$source))) {
+				paste0(x@cpp$source, "::", x@cpp$layer)
 			} else {
-				x@ptr$source
+				x@cpp$source
 			}
 		} else {
 			""
@@ -305,33 +320,35 @@ setMethod("sources", signature(x="SpatVector"),
 
 setMethod("sources", signature(x="SpatVectorProxy"),
 	function(x) {
-		if (x@ptr$v$layer != tools::file_path_sans_ext(basename(x@ptr$v$source))) {
-			paste0(x@ptr$v$source, "::", x@ptr$v$layer)
+		if (x@cpp$v$layer != tools::file_path_sans_ext(basename(x@cpp$v$source))) {
+			paste0(x@cpp$v$source, "::", x@cpp$v$layer)
 		} else {
-			x@ptr$v$source
+			x@cpp$v$source
 		}
 	}
 )
 
 setMethod("hasMinMax", signature(x="SpatRaster"),
 	function(x) {
-		x@ptr$hasRange
+		x@cpp$hasRange
 	}
 )
 
 setMethod("minmax", signature(x="SpatRaster"),
 	function(x, compute=FALSE) {
-		have <- x@ptr$hasRange
+		have <- x@cpp$hasRange
 		if (!all(have)) {
 			if (compute) {
 				opt <- spatOptions()
-				x@ptr$setRange(opt)
+				x@cpp$setRange(opt, FALSE)
 			} else {
 				warn("minmax", "min and max values not available for all layers. See 'setMinMax' or 'global'")
 			}
 		}
-		r <- rbind(x@ptr$range_min, x@ptr$range_max)
-		r[,!have] <- c(Inf, -Inf)
+		r <- rbind(x@cpp$range_min, x@cpp$range_max)
+		if (!compute) {
+			r[,!have] <- c(Inf, -Inf)
+		}
 		colnames(r) <- names(x)
 		rownames(r) <- c("min", "max")
 		r
@@ -343,49 +360,106 @@ setMethod("setMinMax", signature(x="SpatRaster"),
 	function(x, force=FALSE) {
 		opt <- spatOptions()
 		if (force) {
-			x@ptr$setRange(opt)
+			x@cpp$setRange(opt, TRUE)
 		} else if (!all(hasMinMax(x))) {
-			x@ptr$setRange(opt)
+			x@cpp$setRange(opt, FALSE)
 		}
 		x <- messages(x, "setMinMax")
 	}
 )
 
 
+
 setMethod("compareGeom", signature(x="SpatRaster", y="SpatRaster"),
 	function(x, y, ..., lyrs=FALSE, crs=TRUE, warncrs=FALSE, ext=TRUE, rowcol=TRUE, res=FALSE, stopOnError=TRUE, messages=FALSE) {
-		dots <- list(...)
 		opt <- spatOptions("")
-		res <- x@ptr$compare_geom(y@ptr, lyrs, crs, opt$tolerance, warncrs, ext, rowcol, res)
+		out <- x@cpp$compare_geom(y@cpp, lyrs, crs, opt$tolerance, warncrs, ext, rowcol, res)
 		if (stopOnError) {
 			messages(x, "compareGeom")
 		} else {
 			m <- NULL
-			if (x@ptr$has_warning()) {
-				m <- x@ptr$getWarnings()
+			if (x@cpp$has_warning()) {
+				m <- x@cpp$getWarnings()
 			}
-			if (x@ptr$has_error()) {
-				m <- c(m, x@ptr$getError())
+			if (x@cpp$has_error()) {
+				m <- c(m, x@cpp$getError())
 			}
 			if (!is.null(m) && messages) {
 				message(paste(m, collapse="\n"))
 			}
 		}
-		if (length(dots)>1) {
+		dots <- list(...)
+		if (length(dots) > 0) {
 			for (i in 1:length(dots)) {
-				bool <- x@ptr$compare_geom(dots[[i]]@ptr, lyrs, crs, warncrs, ext, rowcol, res)
-				if (stopOnError) messages(x, "compareGeom")
-				res <- bool & res
+				if (!inherits(dots[[i]], "SpatRaster")) {
+					error("compareGeom", "all additional arguments must be a SpatRaster")
+				}
+				out <- out & 
+					compareGeom(x, dots[[i]], lyrs=lyrs, crs=crs, warncrs=warncrs, ext=ext, rowcol=rowcol, res=res, stopOnError=stopOnError, messages=messages)
 			}
 		}
-		res
+		out
+	}
+)
+
+
+setMethod("compareGeom", signature(x="SpatRaster", y="list"),
+	function(x, y, ..., lyrs=FALSE, crs=TRUE, warncrs=FALSE, ext=TRUE, rowcol=TRUE, res=FALSE, stopOnError=TRUE, messages=FALSE) {
+		dots <- list(...)
+		if (length(dots) > 0) {
+			y <- c(y, dots)
+		}
+		isr <- sapply(y, inherits, "SpatRaster")
+		if (!all(isr)) {
+			y <- y[isr]
+			if (length(y) == 0) error("compareGeom", "none of the elements of y is a SpatRaster")
+			n <- sum(!isr)
+			if (n > 1) {
+				warn("compareGeom", paste(n, "elements of y are not a SpatRaster"))
+			} else {
+				warn("compareGeom", paste("1 element of y is not a SpatRaster"))
+			}
+		}
+		out <- sapply(y, compareGeom, y=x, lyrs=lyrs, crs=crs, warncrs=warncrs, ext=ext, rowcol=rowcol, res=res, stopOnError=stopOnError, messages=messages)
+		if (!all(isr)) {
+			res <- rep(NA, length(isr))
+			res[isr] <- out
+			return(res)
+		}
+		all(out)
+	}
+)
+
+
+setMethod("compareGeom", signature(x="SpatRasterCollection", y="missing"),
+	function(x, y, ..., lyrs=FALSE, crs=TRUE, warncrs=FALSE, ext=TRUE, rowcol=TRUE, res=FALSE, stopOnError=FALSE, messages=FALSE) {
+		x <- as.list(x)
+		dots <- list(...)
+		if (length(dots) > 0) {
+			error("compareGeom", "when x is a SpatRasterCollection, additional arguments are ignored")
+		}
+		if (length(x) == 1) return(TRUE)
+		out <- sapply(x[-1], compareGeom, y=x[1], lyrs=lyrs, crs=crs, warncrs=warncrs, ext=ext, rowcol=rowcol, res=res, stopOnError=stopOnError, messages=messages)
+		out
+	}
+)
+
+
+setMethod("compareGeom", signature(x="SpatRaster", y="SpatRasterCollection"),
+	function(x, y, ..., lyrs=FALSE, crs=TRUE, warncrs=FALSE, ext=TRUE, rowcol=TRUE, res=FALSE, stopOnError=FALSE, messages=FALSE) {
+		dots <- list(...)
+		if (length(dots) > 0) {
+			error("compareGeom", "when y is a SpatRasterCollection, additional arguments are ignored")
+		}
+		out <- sapply(as.list(y), compareGeom, y=x, lyrs=lyrs, crs=crs, warncrs=warncrs, ext=ext, rowcol=rowcol, res=res, stopOnError=stopOnError, messages=messages)
+		all(out)
 	}
 )
 
 
 setMethod("compareGeom", signature(x="SpatVector", y="SpatVector"),
 	function(x, y, tolerance=0) {
-		out <- x@ptr$equals_between(y@ptr, tolerance)
+		out <- x@cpp$equals_between(y@cpp, tolerance)
 		x <- messages(x, "compareGeom")
 		out[out == 2] <- NA
 		matrix(as.logical(out), nrow=nrow(x), byrow=TRUE)
@@ -394,7 +468,7 @@ setMethod("compareGeom", signature(x="SpatVector", y="SpatVector"),
 
 setMethod("compareGeom", signature(x="SpatVector", y="SpatVector"),
 	function(x, y, tolerance=0) {
-		out <- x@ptr$equals_within(tolerance)
+		out <- x@cpp$equals_within(tolerance)
 		x <- messages(x, "compareGeom")
 		out[out == 2] <- NA
 		out <- matrix(as.logical(out), nrow=nrow(x), byrow=TRUE)
@@ -403,6 +477,28 @@ setMethod("compareGeom", signature(x="SpatVector", y="SpatVector"),
 )
 
 
+setMethod("all.equal", signature(target="SpatRaster", current="SpatRaster"),
+	function(target, current, maxcell=10000, ...) {
+		a <- base::all.equal(rast(target), rast(current))
+		if (isTRUE(a) && maxcell > 0) {
+			hvT <- hasValues(target)
+			hvC <- hasValues(current)
+			if (hvT && hvC) {
+				s1 <- spatSample(current, maxcell, "regular")
+				s2 <- spatSample(target, maxcell, "regular")
+				a <- all.equal(s1, s2, ...)
+			} else if (hvT || hvC) {
+				if (hvT) {
+					a <- "target has cell values, current does not"
+				} else {
+					a <- "current has cell values, target does not"				
+				}
+			}
+		}
+		a
+	}
+)
+
 
 
 setMethod("values", signature("SpatVector"),
@@ -414,13 +510,13 @@ setMethod("values", signature("SpatVector"),
 
 setMethod("values<-", signature("SpatVector", "data.frame"),
 	function(x, value) {
-		x@ptr <- x@ptr$deepcopy()
+		x@cpp <- x@cpp$deepcopy()
 		if (ncol(value) == 0) {
-			x@ptr$remove_df()
+			x@cpp$remove_df()
 			return(x)
 		}
 		value <- .makeSpatDF(value)
-		x@ptr$set_df(value)
+		x@cpp$set_df(value)
 		messages(x)
 	}
 )
@@ -447,15 +543,15 @@ setMethod("values<-", signature("SpatVector", "ANY"),
 
 setMethod("values<-", signature("SpatVector", "NULL"),
 	function(x, value) {
-		x@ptr <- x@ptr$deepcopy()
-		x@ptr$remove_df()
+		x@cpp <- x@cpp$deepcopy()
+		x@cpp$remove_df()
 		x
 	}
 )
 
 setMethod("setValues", signature("SpatVector"),
 	function(x, values) {
-		x@ptr <- x@ptr$deepcopy()
+		x@cpp <- x@cpp$deepcopy()
 		`values<-`(x, values)
 	}
 )
