@@ -5,7 +5,7 @@
 #include <memory> //std::addressof
 #include "NA.h"
 #include "spatTime.h"
-#include "spatVector2.h"
+//#include "spatVector2.h"
 
 //static void SpatRaster_finalizer( SpatRaster* ptr ){
 //}
@@ -39,36 +39,42 @@ Rcpp::List getDataFrame(SpatDataFrame* v) {
 	if (n == 0) {
 		return(out);
 	}
-
+	long longNA = NA<long>::value;
+	std::string stringNA = v->NAS;
+	SpatTime_t timeNA = NA<SpatTime_t>::value;
+	
 	std::vector<std::string> nms = v->names;
 	std::vector<unsigned> itype = v->itype;
 	for (size_t i=0; i < n; i++) {
 		if (itype[i] == 0) {
 			out[i] = v->getD(i);
 		} else if (itype[i] == 1) {
-			Rcpp::NumericVector iv = Rcpp::wrap(v->getI(i));
-			long longNA = NA<long>::value;
-			for (R_xlen_t j=0; j<iv.size(); j++) {
-				if (iv[j] == longNA) {
-					iv[j] = NA_REAL;
+			//gives warning with NA #1031
+			// Rcpp::wrap(v->getI(i));
+			std::vector<long> ints = v->getI(i);
+			size_t n = ints.size();
+			Rcpp::IntegerVector iv(n, NA_INTEGER);
+			for (size_t j=0; j<n; j++) {
+				if (ints[j] != longNA) {
+					iv[j] = ints[j]; 
 				}
 			}
 			out[i] = iv;
 
 		} else if (itype[i] == 2){
-			Rcpp::CharacterVector s = Rcpp::wrap(v->getS(i));
+			Rcpp::StringVector s = Rcpp::wrap(v->getS(i));
 			for (R_xlen_t j=0; j<s.size(); j++) {
-				if (s[j] == "____NA_+") {
+				if (s[j] == stringNA) {
 					s[j] = NA_STRING;
 				}
 			}
 			out[i] = s;
 		} else if (itype[i] == 3){
 			std::vector<int8_t> b = v->getB(i);
-			Rcpp::NumericVector d(b.size());
+			Rcpp::LogicalVector d(b.size());
 			for (size_t j=0; j<b.size(); j++) {
 				if (b[j] > 1) {
-					d[j] = NA_REAL;
+					d[j] = NA_LOGICAL;
 				} else {
 					d[j] = b[j];
 				}
@@ -77,7 +83,6 @@ Rcpp::List getDataFrame(SpatDataFrame* v) {
 		} else if (itype[i] == 4){
 			SpatTime_v tx = v->getT(i);
 			Rcpp::NumericVector tv = Rcpp::wrap(tx.x);
-			SpatTime_t timeNA = NA<SpatTime_t>::value;
 			for (R_xlen_t j=0; j<tv.size(); j++) {
 				if (tv[j] == timeNA) {
 					tv[j] = NA_REAL;
@@ -181,12 +186,13 @@ RCPP_EXPOSED_CLASS(SpatVector)
 RCPP_EXPOSED_CLASS(SpatVectorProxy)
 RCPP_EXPOSED_CLASS(SpatVectorCollection)
 RCPP_EXPOSED_CLASS(SpatGraph)
-RCPP_EXPOSED_CLASS(SpatVector2)
+//RCPP_EXPOSED_CLASS(SpatVector2)
 
 RCPP_MODULE(spat){
 
     using namespace Rcpp;
 
+/*
     class_<SpatVector2>("SpatVector2")
 	
 		.constructor()
@@ -199,6 +205,7 @@ RCPP_MODULE(spat){
 		.method("from_old", &SpatVector2::from_old)
 		.method("to_old", &SpatVector2::to_old)
 	;
+*/
 
     class_<SpatTime_v>("SpatTime_v")
 		.constructor()
@@ -209,9 +216,11 @@ RCPP_MODULE(spat){
 
     class_<SpatFactor>("SpatFactor")
 		.constructor()
-		.constructor<std::vector<unsigned>, std::vector<std::string>>()
+		.constructor<std::vector<unsigned>, std::vector<std::string>, bool>()
+//		.constructor<std::vector<unsigned>, std::vector<std::string>>()
 		.field("values", &SpatFactor::v)
 		.field("labels", &SpatFactor::labels)
+		.field("ordered", &SpatFactor::ordered)
 	;
 
 
@@ -232,6 +241,8 @@ RCPP_MODULE(spat){
 		.method("deepcopy", &SpatExtent::deepCopy, "deepCopy")
 		.property("vector", &SpatExtent::asVector)
 		.property("valid", &SpatExtent::valid)
+		.property("valid_notempty", &SpatExtent::valid_notempty)
+		.property("empty", &SpatExtent::empty)
 		.method("align", &SpatExtent::align, "align")
 		.method("intersect", &SpatExtent::intersect, "intersect")
 		.method("as.points", &SpatExtent::asPoints, "as.points")
@@ -355,6 +366,7 @@ RCPP_MODULE(spat){
 
     class_<SpatVectorCollection>("SpatVectorCollection")
 		.constructor()
+		.constructor<std::string, std::string, std::string, std::vector<double>, SpatVector>()
 
 		//.property("names", &SpatVectorCollection::get_names, &SpatVectorCollection::set_names)
 		.method("deepcopy", &SpatVectorCollection::deepCopy, "deepCopy")
@@ -499,6 +511,7 @@ RCPP_MODULE(spat){
 		.method("subset_rows", ( SpatVector (SpatVector::*)(std::vector<int>))( &SpatVector::subset_rows ))
 		.method("remove_rows", &SpatVector::remove_rows)
 		.method("type", &SpatVector::type)
+		.method("naGeoms", &SpatVector::naGeoms)
 		.method("nullGeoms", &SpatVector::nullGeoms)
 
 		.method("write", &SpatVector::write)
@@ -515,6 +528,8 @@ RCPP_MODULE(spat){
 
 		.method("disaggregate", &SpatVector::disaggregate)
 		.method("buffer", &SpatVector::buffer)
+		.method("buffer2", &SpatVector::buffer2)
+		.method("buffer3", &SpatVector::buffer3)
 		.method("centroid", &SpatVector::centroid)
 		.method("point_on_surface", &SpatVector::point_on_surface)
 		.method("make_valid2", &SpatVector::make_valid2)
@@ -555,7 +570,7 @@ RCPP_MODULE(spat){
 		.method("equals_within", ( std::vector<unsigned> (SpatVector::*)(bool, double))( &SpatVector::equals_exact ))
 
 
-		.method("crop_ext", ( SpatVector (SpatVector::*)(SpatExtent))( &SpatVector::crop ))
+		.method("crop_ext", ( SpatVector (SpatVector::*)(SpatExtent, bool))( &SpatVector::crop ))
 		.method("crop_vct", ( SpatVector (SpatVector::*)(SpatVector))( &SpatVector::crop ))
 
 		.method("near_between", (SpatVector (SpatVector::*)(SpatVector, bool))( &SpatVector::nearest_point))
@@ -575,6 +590,7 @@ RCPP_MODULE(spat){
 
 		.method("densify", &SpatVector::densify)
 		.method("round", &SpatVector::round)
+		.method("make_CCW", &SpatVector::make_CCW)
 	;
 
 
@@ -616,6 +632,7 @@ RCPP_MODULE(spat){
 		.constructor()
 	 // .constructor<std::string, int>()
 	    .constructor<std::vector<std::string>, std::vector<int>, std::vector<std::string>, bool, std::vector<std::string>, std::vector<std::string>, std::vector<size_t>>()
+		
 		.constructor<std::vector<unsigned>, std::vector<double>, std::string>()
 		//.finalizer(&SpatRaster_finalizer)
 
@@ -626,6 +643,12 @@ RCPP_MODULE(spat){
 		.method("getWarnings", &SpatRaster::getWarnings)
 		.method("getMessage", &SpatRaster::getMessage)
 
+		.method("addTag", &SpatRaster::addTag)
+		.method("getTags", &SpatRaster::getTags)
+		.method("addLyrTags", &SpatRaster::addLyrTags)
+		.method("getLyrTags", &SpatRaster::getLyrTags)
+
+
 		//.field("name", &SpatRaster::name)
 		.method("getFileBlocksize", &SpatRaster::getFileBlocksize)
 
@@ -644,6 +667,8 @@ RCPP_MODULE(spat){
 		//.field_readonly("prj", &SpatRaster::prj)
 		.property("extent", &SpatRaster::getExtent, &SpatRaster::setExtent )
 
+		.method("is_rotated", &SpatRaster::is_rotated)
+
 		.method("setWindow", &SpatRaster::setWindow, "")
 		.method("removeWindow", &SpatRaster::removeWindow, "")
 		.method("hasWindow", &SpatRaster::hasWindow, "")
@@ -710,6 +735,7 @@ RCPP_MODULE(spat){
 		.property("timezone", &SpatRaster::getTimeZone)
 		.method("settime", &SpatRaster::setTime)
 		//.property("timestr", &SpatRaster::getTimeStr)
+		.method("metadata", &SpatRaster::getMetadata)
 
 		.property("depth", &SpatRaster::getDepth)
 		.method("set_depth", &SpatRaster::setDepth)
@@ -744,8 +770,9 @@ RCPP_MODULE(spat){
 		.method("setUnit", &SpatRaster::setUnit)
 		.method("set_resolution", &SpatRaster::setResolution)
 		.method("subset", &SpatRaster::subset)
+		.method("subsetSource", &SpatRaster::subsetSource)
 
-		.method("cellFromXY", ( std::vector<double> (SpatRaster::*)(std::vector<double>,std::vector<double>) )( &SpatRaster::cellFromXY ))
+		.method("cellFromXY", ( std::vector<double> (SpatRaster::*)(std::vector<double>,std::vector<double>, double) )( &SpatRaster::cellFromXY ))
 		.method("vectCells", &SpatRaster::vectCells)
 		.method("extCells", &SpatRaster::extCells)
 
@@ -759,6 +786,7 @@ RCPP_MODULE(spat){
 		.method("colFromX", ( std::vector<int_64> (SpatRaster::*)(const std::vector<double>&) )( &SpatRaster::colFromX ))
 		.method("rowFromY", ( std::vector<int_64> (SpatRaster::*)(const std::vector<double>&) )( &SpatRaster::rowFromY ))
 		.method("xyFromCell", ( std::vector< std::vector<double> > (SpatRaster::*)(std::vector<double>&) )( &SpatRaster::xyFromCell ))
+		.method("crds", &SpatRaster::coordinates)
 		.method("rowColFromCell", ( std::vector< std::vector<int_64> > (SpatRaster::*)(std::vector<double>) )( &SpatRaster::rowColFromCell ))
 		.method("readStart", &SpatRaster::readStart)
 		.method("readStop", &SpatRaster::readStop)
@@ -794,13 +822,13 @@ RCPP_MODULE(spat){
 		.method("rappvals", &SpatRaster::rappvals)
 		.method("roll", &SpatRaster::roll)
 		.method("fill_range", &SpatRaster::fill_range)
-		.method("arith_rast", (SpatRaster (SpatRaster::*)(SpatRaster, std::string, SpatOptions&) )( &SpatRaster::arith ))
-		.method("arith_numb", (SpatRaster (SpatRaster::*)(std::vector<double>, std::string, bool, SpatOptions&) )( &SpatRaster::arith))
+		.method("arith_rast", (SpatRaster (SpatRaster::*)(SpatRaster, std::string, bool, SpatOptions&) )( &SpatRaster::arith ))
+		.method("arith_numb", (SpatRaster (SpatRaster::*)(std::vector<double>, std::string, bool, bool,SpatOptions&) )( &SpatRaster::arith))
 		.method("arith_m", &SpatRaster::arith_m)
 		
 		.method("rst_area", &SpatRaster::rst_area)
 		.method("sum_area", &SpatRaster::sum_area)
-		.method("area_by_value", &SpatRaster::area_by_value)
+		.method("sum_area_group", &SpatRaster::sum_area_group)
 
 		.method("as_points", &SpatRaster::as_points)
 		.method("as_points_value", &SpatRaster::as_points_value)
@@ -833,6 +861,7 @@ RCPP_MODULE(spat){
 		.method("vectDistanceRasterize", &SpatRaster::distance_rasterize)
 		.method("vectDirectionRasterize", &SpatRaster::direction_rasterize)
 		.method("clamp", &SpatRaster::clamp)
+		.method("clamp_raster", &SpatRaster::clamp_raster)
 		.method("clamp_ts", &SpatRaster::clamp_ts)
 		.method("replaceValues", &SpatRaster::replaceValues)
 		.method("classify", ( SpatRaster (SpatRaster::*)(std::vector<double>, unsigned, unsigned, bool, bool, double, bool, bool, bool, SpatOptions&) )( &SpatRaster::reclassify))
@@ -862,7 +891,9 @@ RCPP_MODULE(spat){
 
 		.method("get_aggregates", &SpatRaster::get_aggregates)
 		.method("get_aggregate_dims", &SpatRaster::get_aggregate_dims2)
-		.method("global", &SpatRaster::global)
+	//	.method("global", &SpatRaster::global)
+		.method("mglobal", &SpatRaster::mglobal)
+		.method("layerCor", &SpatRaster::layerCor)
 		.method("global_weighted_mean", &SpatRaster::global_weighted_mean)
 
 		.method("initf", ( SpatRaster (SpatRaster::*)(std::string, bool, SpatOptions&) )( &SpatRaster::init ), "init fun")
@@ -873,19 +904,28 @@ RCPP_MODULE(spat){
 		.method("nonan", &SpatRaster::nonan)
 		.method("allnan", &SpatRaster::allnan)
 		.method("isnan", &SpatRaster::isnan)
+		.method("countnan", &SpatRaster::countnan)
 		.method("not_na", &SpatRaster::isnotnan)
 		.method("isfinite", &SpatRaster::isfinite)
 		.method("isinfinite", &SpatRaster::isinfinite)
+		.method("is_true", &SpatRaster::is_true)
+		.method("is_false", &SpatRaster::is_false)
+
 		.method("logic_rast", ( SpatRaster (SpatRaster::*)(SpatRaster, std::string, SpatOptions&) )( &SpatRaster::logic ))
-		.method("logic_numb", ( SpatRaster (SpatRaster::*)(bool, std::string, SpatOptions&) )( &SpatRaster::logic ))
-		.method("mask_raster", ( SpatRaster (SpatRaster::*)(SpatRaster, bool, std::vector<double>, double, SpatOptions&) )( &SpatRaster::mask))
-		.method("mask_vector", ( SpatRaster (SpatRaster::*)(SpatVector, bool, double, bool, SpatOptions&) )( &SpatRaster::mask))
+		.method("logic_numb", ( SpatRaster (SpatRaster::*)(std::vector<double>, std::string, SpatOptions&) )( &SpatRaster::logic ))
+		.method("mask_self", ( SpatRaster (SpatRaster::*)(SpatOptions&) )( &SpatRaster::mask))
+		.method("mask_raster", ( SpatRaster (SpatRaster::*)(SpatRaster&, bool, std::vector<double>, double, SpatOptions&) )( &SpatRaster::mask))
+		.method("mask_vector", ( SpatRaster (SpatRaster::*)(SpatVector&, bool, double, bool, SpatOptions&) )( &SpatRaster::mask))
 		.method("math", &SpatRaster::math)
 		.method("math2", &SpatRaster::math2)
 		.method("modal", &SpatRaster::modal)
 		.method("quantile", &SpatRaster::quantile)
 		.method("rasterize", &SpatRaster::rasterize)
-		.method("rasterizePoints", &SpatRaster::rasterizePoints)
+		
+		.method("rasterizePointsV", ( SpatRaster (SpatRaster::*)(SpatVector&, std::string, std::vector<double>&, bool, double, SpatOptions&) )( &SpatRaster::rasterizePoints))
+
+		.method("rasterizePointsXY", ( SpatRaster (SpatRaster::*)(std::vector<double>&, std::vector<double>&, std::string, std::vector<double>&, bool, double, SpatOptions&) )( &SpatRaster::rasterizePoints))
+
 		.method("rasterizeLyr", &SpatRaster::rasterizeLyr)
 		.method("rasterizeGeom", &SpatRaster::rasterizeGeom)
 		.method("rasterizeWindow", &SpatRaster::rasterizeWindow)
@@ -906,6 +946,7 @@ RCPP_MODULE(spat){
 		.method("sampleRandomValues", &SpatRaster::sampleRandomValues)
 		.method("scale", &SpatRaster::scale)
 		.method("shift", &SpatRaster::shift)
+		.method("similarity", &SpatRaster::similarity)
 		.method("terrain", &SpatRaster::terrain)
 		.method("hillshade", &SpatRaster::hillshade)
 		.method("summary", &SpatRaster::summary)
@@ -923,19 +964,19 @@ RCPP_MODULE(spat){
 		.method("rectify", &SpatRaster::rectify)
 		.method("stretch", &SpatRaster::stretch)
 		.method("warp", &SpatRaster::warper)
+    .method("warp_by_util", &SpatRaster::warper_by_util)
 		.method("resample", &SpatRaster::resample)
 		.method("zonal", &SpatRaster::zonal)
 		.method("zonal_weighted", &SpatRaster::zonal_weighted)
 		.method("zonal_poly", &SpatRaster::zonal_poly)		
+		.method("zonal_poly_table", &SpatRaster::zonal_poly_table)		
 		.method("zonal_poly_weighted", &SpatRaster::zonal_poly_weighted)		
 //		.method("zonal_old", &SpatRaster::zonal_old)
-		.method("is_true", &SpatRaster::is_true)
-		.method("is_false", &SpatRaster::is_false)
 	;
 
     class_<SpatRasterCollection>("SpatRasterCollection")
 		.constructor()
-	    .constructor<std::string, std::vector<int>, bool>()
+	    .constructor<std::string, std::vector<int>, bool, std::vector<std::string>>()
 
 		.property("names", &SpatRasterCollection::get_names, &SpatRasterCollection::set_names)
 
@@ -956,11 +997,12 @@ RCPP_MODULE(spat){
 		.method("merge", &SpatRasterCollection::merge)
 		.method("mosaic", &SpatRasterCollection::mosaic)
 		.method("morph", &SpatRasterCollection::morph)
+		.method("crop", &SpatRasterCollection::crop)
 	;
 
     class_<SpatRasterStack>("SpatRasterStack")
 		.constructor()
-	    .constructor<std::string, std::vector<int>, bool>()
+	    .constructor<std::string, std::vector<int>, bool, std::vector<std::string>>()
 	    .constructor<SpatRaster, std::string, std::string, std::string>()
 		.method("deepcopy", &SpatRasterStack::deepCopy)
 
