@@ -16,7 +16,7 @@ void getGDALdriver(std::string &filename, std::string &driver) {
 	lrtrim(filename);
 	lrtrim(driver);
 
-	if (driver != "") {
+	if (!driver.empty()) {
 		if (driver == "RST") {
 			filename = noext(filename) + ".rst";
 		}
@@ -54,7 +54,7 @@ void getGDALdriver(std::string &filename, std::string &driver) {
 bool SpatRaster::getTempFile(std::string &filename, std::string &driver, SpatOptions& opt) {
 
 	driver = opt.get_def_filetype();
-	if ((driver == "") || (driver == "GTiff")) {
+	if (driver.empty() || (driver == "GTiff")) {
 		driver = "GTiff";
 		filename = tempFile(opt.get_tempdir(), opt.pid, ".tif");
 		return true;
@@ -177,7 +177,7 @@ std::vector<std::vector<std::string>> parse_metadata_sds(std::vector<std::string
 			size_t pos = s.find(ddelim);
 			if (pos != std::string::npos) {
 				s.erase(0, pos + ddelim.length());
-				pos = s.find("]");
+				pos = s.find(']');
 				std::string dims = s.substr(1, pos-1);
 
 				std::vector<std::string> d = strsplit(dims, "x");
@@ -193,9 +193,9 @@ std::vector<std::vector<std::string>> parse_metadata_sds(std::vector<std::string
 					size_t ds = d.size()-1;
 					size_t nls = 0;
 					try {
-						nls = stoi(d[ds-2]);
+						nls = std::stol(d[ds-2]);
 						for (size_t i=0; i<(ds-2); i++) {
-							nls *= stoi(d[i]);
+							nls *= std::stol(d[i]);
 						}
 					} catch(...) {}
 					nl.push_back(std::to_string(nls));
@@ -204,7 +204,7 @@ std::vector<std::vector<std::string>> parse_metadata_sds(std::vector<std::string
 				}
 				//desc.push_back(std::string(pos, s.size()));
 				s = s.substr(pos+2, s.size());
-				pos = s.find(" ");
+				pos = s.find(' ');
 				s = s.substr(0, pos);
 				desc.push_back(s);
 
@@ -254,7 +254,7 @@ std::vector<std::vector<std::string>> sdinfo(std::string fname) {
 	for (size_t i=0; metadata[i] != NULL; i++) {
 		meta.push_back(metadata[i]);
 	}
-	if (meta.size() == 0) {
+	if (meta.empty()) {
 		GDALClose( (GDALDatasetH) poDataset );
 		out[0] = std::vector<std::string> {"no subdatasets"};
 		return out;
@@ -311,10 +311,9 @@ std::vector<std::vector<std::string>> sdinfo(std::string fname) {
 
 #if GDAL_VERSION_MAJOR <= 2 && GDAL_VERSION_MINOR < 1
 
-SpatRaster SpatRaster::make_vrt(std::vector<std::string> filenames, std::vector<std::string> options, SpatOptions &opt) {
-	SpatRaster out;
-	out.setError( "GDAL version >= 2.1 required for vrt");
-	return out;
+std::string SpatRaster::make_vrt(std::vector<std::string> filenames, std::vector<std::string> options, SpatOptions &opt) {
+	setError( "GDAL version >= 2.1 required for vrt");
+	return("");
 }
 
 
@@ -328,15 +327,14 @@ std::string gdalinfo(std::string filename, std::vector<std::string> options, std
 
 # include "gdal_utils.h" // requires >= 2.1
 
-SpatRaster SpatRaster::make_vrt(std::vector<std::string> filenames, std::vector<std::string> options, SpatOptions &opt) {
+std::string SpatRaster::make_vrt(std::vector<std::string> filenames, std::vector<std::string> options, SpatOptions &opt) {
 
-	SpatRaster out;
 	std::string outfile = opt.get_filename();
-	if (outfile == "") {
+	if (outfile.empty()) {
 		outfile = tempFile(opt.get_tempdir(), opt.pid, ".vrt");
 	} else if (file_exists(outfile) && (!opt.get_overwrite())) {
-		out.setError("output file exists. You can use 'overwrite=TRUE' to overwrite it");
-		return(out);
+		setError("output file exists. You can use 'overwrite=TRUE' to overwrite it");
+		return("");
 	}
 
 /*
@@ -377,9 +375,9 @@ SpatRaster SpatRaster::make_vrt(std::vector<std::string> filenames, std::vector<
 	std::vector <char *> vops = string_to_charpnt(options);
 	GDALBuildVRTOptions* vrtops = GDALBuildVRTOptionsNew(vops.data(), NULL);
 	if (vrtops == NULL) {
-		out.setError("options error");
+		setError("options error");
 		CSLDestroy( names );
-		return(out);
+		return("");
 	}
 	int pbUsageError;
 //	GDALDataset *ds = (GDALDataset *) GDALBuildVRT(outfile.c_str(), tiles.size(), (GDALDatasetH *) tiles.data(), nullptr, vrtops, &pbUsageError);
@@ -391,15 +389,11 @@ SpatRaster SpatRaster::make_vrt(std::vector<std::string> filenames, std::vector<
 
 	//for (size_t i= 0; i<tiles.size(); i++) GDALClose(tiles[i]);
 	if(ds == NULL )  {
-		out.setError("cannot create vrt. Error #"+ std::to_string(pbUsageError));
-		return out;
+		setError("cannot create vrt. Error #"+ std::to_string(pbUsageError));
+		return("");
 	}
 	GDALClose(ds);
-	if (!out.constructFromFile(outfile, {-1}, {""}, {}, {})) {
-		out.setError("cannot open created vrt");
-		return out;
-	}
-	return out;
+	return(outfile);
 }
 
 
@@ -761,7 +755,7 @@ bool SpatRaster::from_gdalMEM(GDALDatasetH hDS, bool set_geometry, bool get_valu
 		if (!s.srs.set({wkt}, msg)) {
 			setError(msg);
 			return false;
-		} else if (msg != "") {
+		} else if (!msg.empty()) {
 			addWarning(msg);
 		}
 
@@ -799,6 +793,7 @@ bool SpatRaster::from_gdalMEM(GDALDatasetH hDS, bool set_geometry, bool get_valu
 			}
 
 
+// mem should never has scale/offset? 
 			bool has_so = false;;
 			int success;
 			double moffset = GDALGetRasterOffset(hBand, &success);
@@ -965,7 +960,7 @@ bool SpatRaster::create_gdalDS(GDALDatasetH &hDS, std::string filename, std::str
 	GDALSetGeoTransform( hDS, adfGeoTransform);
 
 	std::string wkt = getSRS("wkt");
-	if (wkt != "") {
+	if (!wkt.empty()) {
 		OGRSpatialReferenceH hSRS = OSRNewSpatialReference( NULL );
 		OGRErr erro = OSRSetFromUserInput(hSRS, wkt.c_str());
 		if (erro == 4) {
