@@ -22,6 +22,7 @@
 #include "recycle.h"
 #include "string_utils.h"
 
+
 void callbck(void *item, void *userdata) { // callback function for tree selection
 	std::vector<size_t> *ret = (std::vector<size_t> *) userdata;
 	ret->push_back(*((size_t *) item));
@@ -43,7 +44,7 @@ SpatVectorCollection SpatVector::bienvenue() {
 	return out;
 }
 
-
+/*
 std::vector<std::string> SpatVector::wkt() {
 	GEOSContextHandle_t hGEOSCtxt = geos_init();
 	std::vector<GeomPtr> g = geos_geoms(this, hGEOSCtxt);
@@ -57,6 +58,22 @@ std::vector<std::string> SpatVector::wkt() {
 	geos_finish(hGEOSCtxt);
 	return out;
 }
+*/
+
+std::vector<std::string> SpatVector::wkt() {
+	GEOSContextHandle_t hGEOSCtxt = geos_init();
+	std::vector<GeomPtr> g = geos_geoms(this, hGEOSCtxt);
+	std::vector<std::string> out;
+	out.reserve(g.size());
+	GEOSWKTWriter* writer = GEOSWKTWriter_create_r(hGEOSCtxt);
+	for (size_t i = 0; i < g.size(); i++) {
+		char *wkt = GEOSGeomToWKT_r(hGEOSCtxt, g[i].get());
+		out.push_back(wkt);
+	}
+	GEOSWKTWriter_destroy_r(hGEOSCtxt, writer);
+	geos_finish(hGEOSCtxt);
+	return out;
+}
 
 
 
@@ -65,17 +82,21 @@ std::vector<std::string> SpatVector::wkb() {
 	std::vector<GeomPtr> g = geos_geoms(this, hGEOSCtxt);
 	std::vector<std::string> out;
 	out.reserve(g.size());
-	size_t len = 0;
-	for (size_t i = 0; i < g.size(); i++) {
-		unsigned char *wkb = GEOSGeomToWKB_buf_r(hGEOSCtxt, g[i].get(), &len);
-		std::string s( reinterpret_cast<char const*>(wkb), len) ;
+	GEOSWKBWriter* writer = GEOSWKBWriter_create_r(hGEOSCtxt);
+	size_t len=0;
+	for (size_t i=0; i<g.size(); i++) {
+		unsigned char *wkb = GEOSWKBWriter_write_r(hGEOSCtxt, writer, g[i].get(), &len);
+		std::string s(reinterpret_cast<char const*>(wkb), len);
 		out.push_back(s);
-		free(wkb);
+		GEOSFree_r(hGEOSCtxt, wkb);
 	}
+	GEOSWKBWriter_destroy_r(hGEOSCtxt, writer);
 	geos_finish(hGEOSCtxt);
 	return out;
 }
 
+	
+	
 std::vector<std::string> SpatVector::hex() {
 	GEOSContextHandle_t hGEOSCtxt = geos_init();
 	std::vector<GeomPtr> g = geos_geoms(this, hGEOSCtxt);
@@ -93,7 +114,6 @@ std::vector<std::string> SpatVector::hex() {
 }
 
 
-
 SpatVector SpatVector::from_hex(std::vector<std::string> x, std::string srs) {
 	GEOSContextHandle_t hGEOSCtxt = geos_init();
 	size_t n = x.size();
@@ -117,6 +137,7 @@ SpatVector SpatVector::from_hex(std::vector<std::string> x, std::string srs) {
 }
 
 
+
 SpatVectorCollection SpatVectorCollection::from_hex_col(std::vector<std::string> x, std::string srs) {
 	GEOSContextHandle_t hGEOSCtxt = geos_init();
 	size_t n = x.size();
@@ -826,8 +847,6 @@ SpatVector SpatVector::snapto(SpatVector y, double tolerance) {
 //Geometry * GEOSPolygonize_full_r(GEOSContextHandle_t extHandle, const Geometry* g, Geometry** cuts, Geometry** dangles, Geometry** invalid)
 
 
-
-
 SpatVector SpatVector::crop(SpatVector v) {
 
 	SpatVector out;
@@ -839,9 +858,11 @@ SpatVector SpatVector::crop(SpatVector v) {
 //	if ((type() != "polygons") & (type() != "mutlipolygons")) {
 	if ((v.type() != "polygons")) {
 		v = v.hull("convex");
-	} else {
+	} else if (v.nrow() > 1) {
 		v = v.aggregate(false);
 	}
+	
+//	return intersect(v, false);
 	std::vector<GeomPtr> y = geos_geoms(&v, hGEOSCtxt);
 	std::vector<GeomPtr> result;
 	std::vector<long> ids;
@@ -849,7 +870,9 @@ SpatVector SpatVector::crop(SpatVector v) {
 	ids.reserve(nx);
 
 	for (size_t i = 0; i < nx; i++) {
+	//	Rcpp::Rcout << "intersection " << i;
 		GEOSGeometry* geom = GEOSIntersection_r(hGEOSCtxt, x[i].get(), y[0].get());
+	//	Rcpp::Rcout << " done" << std::endl;
 		if (geom == NULL) {
 			out.setError("GEOS exception");
 			geos_finish(hGEOSCtxt);
@@ -930,6 +953,13 @@ SpatVector SpatVector::hull(std::string htype, std::string by) {
 	GEOSGeometry* h;
 	if (htype == "convex") {
 		h = GEOSConvexHull_r(hGEOSCtxt, g[0].get());
+	} else if (htype == "circle") {
+	#ifndef GEOS380
+		out.setError("GEOS 3.8 required for bounding circle");
+		return out;
+	#else
+		h = GEOSMinimumBoundingCircle_r(hGEOSCtxt, g[0].get(), NULL, NULL);
+	#endif
 	} else {
 	#ifndef GEOS361
 		out.setError("GEOS 3.6.1 required for rotated rectangle");
@@ -938,6 +968,9 @@ SpatVector SpatVector::hull(std::string htype, std::string by) {
 		h = GEOSMinimumRotatedRectangle_r(hGEOSCtxt, g[0].get());
 	#endif
 	}
+	
+	
+	
 	std::vector<GeomPtr> b(1);
 	b[0] = geos_ptr(h, hGEOSCtxt);
 	SpatVectorCollection coll = coll_from_geos(b, hGEOSCtxt);
@@ -953,6 +986,11 @@ SpatVector SpatVector::hull(std::string htype, std::string by) {
 SpatVector SpatVector::voronoi(SpatVector bnd, double tolerance, int onlyEdges) {
 	SpatVector out;
 
+	if (nrow() == 0) {
+		out.addWarning("input SpatVector has no geometries");
+		return out;
+	}
+
 #ifndef GEOS350
 	out.setError("GEOS 3.5 required for voronoi");
 	return out;
@@ -1012,6 +1050,10 @@ SpatVector SpatVector::voronoi(SpatVector bnd, double tolerance, int onlyEdges)
 
 SpatVector SpatVector::delaunay(double tolerance, int onlyEdges) {
 	SpatVector out;
+	if (nrow() == 0) {
+		out.addWarning("input SpatVector has no geometries");
+		return out;
+	}
 
 #ifndef GEOS350
 	out.setError("GEOS 3.5 required for delaunay");
@@ -1042,96 +1084,15 @@ SpatVector SpatVector::delaunay(double tolerance, int onlyEdges) {
 }
 
 
-SpatGeom hullify(SpatVector b, bool ispoly) {
-	if (b.nrow() == 1) return b.geoms[0];
-	if (ispoly) b.addGeom(b.geoms[0]);
-	SpatVector part;
-	part.reserve(b.size());
-	for (size_t j =0; j<(b.size()-1); j++) {
-		std::vector<unsigned> range = {(unsigned)j, (unsigned)j+1};
-		SpatVector g = b.subset_rows(range);
-		g = g.hull("convex");
-		part.addGeom(g.geoms[0]);
-	}
-	part = part.aggregate(true);
-	return part.geoms[0];
-}
 
-
-SpatVector lonlat_buf(SpatVector x, double dist, unsigned quadsegs, bool ispol, bool ishole) {
-
-
-	if ((x.extent.ymin > -60) && (x.extent.ymax < 60) && ((x.extent.ymax - x.extent.ymin) < 1) && dist < 110000) {
-		SpatSRS insrs = x.srs;
-		x.setSRS("+proj=merc");
-		double f = 0.5 - (dist / 220000);
-		double halfy = x.extent.ymin + f * (x.extent.ymax - x.extent.ymin);
-		std::vector<double> dd = destpoint_lonlat(0, halfy, 0, dist);
-		dist = dd[1] - halfy;
-		if (ishole) dist = -dist;
-		x = x.buffer({dist}, quadsegs, "", "", NAN, false);	
-		x.srs = insrs;
-		return x;
-	}
-
-	x = x.disaggregate(false);
-	SpatVector tmp;
-	tmp.reserve(x.size());
-	for (size_t i =0; i<x.geoms.size(); i++) {
-		SpatVector p(x.geoms[i]);
-		p.srs = x.srs;
-		p = p.as_points(false, true);
-		std::vector<double> d(p.size(), dist);
-		SpatVector b = p.point_buffer(d, quadsegs, true);
-		if (b.size() <= p.size()) {
-			SpatGeom g = hullify(b, ispol);
-			tmp.addGeom(g);
-		} else {
-			SpatVector west, east, eastwest;
-			for (size_t j =0; j<b.size(); j++) {
-				if ((b.geoms[j].extent.xmin < -179.99) && (b.geoms[j].extent.xmax > 179.99)) {
-					tmp.addGeom(b.geoms[j]);
-				} else if (b.geoms[j].extent.xmax < 0) {
-					west.addGeom(b.geoms[j]);
-				} else {
-					east.addGeom(b.geoms[j]);
-				}
-			}
-			if (east.nrow() > 0) {
-				SpatGeom geast = hullify(east, ispol);
-				tmp.addGeom(geast);
-			}
-			if (west.nrow() > 0) {
-				SpatGeom gwest = hullify(west, ispol);
-				tmp.addGeom(gwest);
-			}
-		}
-	}
-	tmp = tmp.aggregate(true);
-
-	if (ispol) {
-		if (dist < 0) {
-			tmp = !ishole ? tmp.get_holes() : tmp.remove_holes();
-		} else {
-			tmp = ishole ? tmp.get_holes() : tmp.remove_holes();
-		}
-	}
-
-	return tmp;
-}
-
-
-//, std::vector<std::string> pars
 SpatVector SpatVector::buffer(std::vector<double> d, unsigned quadsegs, std::string capstyle, std::string joinstyle, double mitrelimit, bool singlesided) {
 
-	quadsegs = std::min(quadsegs, (unsigned) 180);
+//	quadsegs = std::min(quadsegs, (unsigned) 180);
 	SpatVector out;
-//	out.srs = srs;
 	if (srs.is_empty()) {
 		out.addWarning("unknown CRS. Results may be wrong");
 	} 
 	bool islonlat = is_lonlat();
-	
 	if (d.size() == 1 && d[0] == 0) {
 		islonlat = false; //faster
 	}
@@ -1139,46 +1100,17 @@ SpatVector SpatVector::buffer(std::vector<double> d, unsigned quadsegs, std::str
 	if (vt == "points" || vt == "lines") {
 		for (size_t i=0; i<d.size(); i++) {
 			if (d[i] <= 0) {
-				d[i] = -d[i];
+				out.setError("a negative buffer is only meaningful with polygons");
+				return out;
 			}
 		}
 	}
 	recycle(d, size());
-
 	if (islonlat) {
-		if (vt == "points") {
-			return point_buffer(d, quadsegs, false);
-		} else {
-			SpatVector p;
-			bool ispol = vt == "polygons";
-			for (size_t i =0; i<size(); i++) {
-				p = subset_rows(i);
-				if (ispol) {
-					SpatVector h = p.get_holes();
-					p = p.remove_holes();
-					p = lonlat_buf(p, d[i], quadsegs, true, false);
-					if (!h.empty()) {
-						h = lonlat_buf(h, d[i], quadsegs, true, true);
-						if (!h.empty()) {
-							for (size_t j=0; j<h.geoms[0].parts.size(); j++) {
-								p.geoms[0].parts[0].addHole(h.geoms[0].parts[j].x, h.geoms[0].parts[j].y);
-							}
-						}
-					}
-				} else {
-					p = lonlat_buf(p, d[i], quadsegs, false, false);
-				}
-				out = out.append(p, true);
-			}
-			out.df = df;
-			out.srs = srs;
-			return out;
-		}
+		return buffer_lonlat(vt, d, quadsegs);
 	}
 
-
 	GEOSContextHandle_t hGEOSCtxt = geos_init();
-//	SpatVector f = remove_holes();
 
 	GEOSBufferParams* bufparms = GEOSBufferParams_create_r(hGEOSCtxt);	
 	GEOSBufferParams_setQuadrantSegments_r(hGEOSCtxt, bufparms, quadsegs);
@@ -1217,10 +1149,7 @@ SpatVector SpatVector::buffer(std::vector<double> d, unsigned quadsegs, std::str
 	}
 	SpatVectorCollection coll = coll_from_geos(b, hGEOSCtxt);
 
-
 	GEOSBufferParams_destroy_r(hGEOSCtxt, bufparms);	
-
-//	out = spat_from_geom(hGEOSCtxt, g, "points");
 	geos_finish(hGEOSCtxt);
 	out = coll.get(0);
 	out.srs = srs;
@@ -1229,6 +1158,37 @@ SpatVector SpatVector::buffer(std::vector<double> d, unsigned quadsegs, std::str
 }
 
 
+// basic version of buffer, for debugging
+SpatVector SpatVector::buffer2(std::vector<double> d, unsigned quadsegs) {
+
+	SpatVector out;
+	recycle(d, size());
+	GEOSContextHandle_t hGEOSCtxt = geos_init();
+	std::vector<GeomPtr> g = geos_geoms(this, hGEOSCtxt);
+	std::vector<GeomPtr> b(size());
+	for (size_t i = 0; i < g.size(); i++) {
+		
+		Rcpp::Rcout << "buffer " << i;		
+		GEOSGeometry* pt = GEOSBuffer_r(hGEOSCtxt, g[i].get(), d[i], (int) quadsegs);
+		Rcpp::Rcout << " done" << std::endl;
+
+		if (pt == NULL) {
+			out.setError("GEOS exception");
+			geos_finish(hGEOSCtxt);
+			return(out);
+		}
+		b[i] = geos_ptr(pt, hGEOSCtxt);
+	}
+	SpatVectorCollection coll = coll_from_geos(b, hGEOSCtxt);
+	geos_finish(hGEOSCtxt);
+	out = coll.get(0);
+	out.srs = srs;
+	out.df = df;
+	return out;
+}
+
+
+
 SpatVector SpatVector::intersect(SpatVector v, bool values) {
 
 	SpatVector out;
@@ -1293,10 +1253,9 @@ SpatVector SpatVector::intersect(SpatVector v, bool values) {
 			}
 		}
 
-
 	//SpatVectorCollection coll = coll_from_geos(result, hGEOSCtxt);
 		if (!result.empty()) {
-			SpatVectorCollection coll = coll_from_geos(result, hGEOSCtxt, ids, true, false);
+			SpatVectorCollection coll = coll_from_geos(result, hGEOSCtxt, ids, false, false);
 			out = coll.get(0);
 			out.srs = srs;
 		}
@@ -2348,6 +2307,11 @@ SpatVector SpatVector::unite(SpatVector v) {
 		return out;
 	}
 
+	if (gtp != "polygons") {
+		// should make records unique here. This is now done in R
+		return append(v, true);
+	}
+
 	out = intersect(v, true);
 	if (out.hasError()) {
 		return out;
@@ -2355,6 +2319,7 @@ SpatVector SpatVector::unite(SpatVector v) {
 	if (out.nrow() == 0) {
 		return append(v, true);
 	}
+
 	if (gtp != out.type()) {
 		SpatVector empty;
 		out = empty;
@@ -3002,6 +2967,10 @@ SpatVector SpatVector::cross_dateline(bool &fixed) {
 SpatVector SpatVector::centroid(bool check_lonlat) {
 
 	SpatVector out;
+	if (nrow() == 0) {
+		out.setError("input has no geometries");
+		return out;
+	}
 
 	if (check_lonlat && could_be_lonlat()) {
 		bool changed = false;
@@ -3037,6 +3006,10 @@ SpatVector SpatVector::centroid(bool check_lonlat) {
 SpatVector SpatVector::point_on_surface(bool check_lonlat) {
 
 	SpatVector out;
+	if (nrow() == 0) {
+		out.setError("input has no geometries");
+		return out;
+	}
 
 	if (check_lonlat && could_be_lonlat()) {
 		bool changed = false;
@@ -3167,3 +3140,53 @@ SpatVector SpatVector::clearance() {
 }
 
 
+bool SpatPart::is_CCW() {
+#ifndef GEOS370
+	return true;
+#else
+	GEOSContextHandle_t hGEOSCtxt = geos_init();
+	GEOSCoordSequence *pseq;
+	size_t n = size();
+	pseq = GEOSCoordSeq_create_r(hGEOSCtxt, n, 2);
+	for (size_t i = 0; i < n; i++) {
+		GEOSCoordSeq_setX_r(hGEOSCtxt, pseq, i, x[i]);
+		GEOSCoordSeq_setY_r(hGEOSCtxt, pseq, i, y[i]);
+	}
+	char is_ccw;
+	bool success = GEOSCoordSeq_isCCW_r(hGEOSCtxt, pseq, &is_ccw);
+	GEOSCoordSeq_destroy_r(hGEOSCtxt, pseq);
+	geos_finish(hGEOSCtxt);
+	if (success) {
+		return is_ccw != 0;
+	} else {
+		return true;
+	}
+#endif
+}
+
+
+
+void SpatVector::make_CCW() {
+	#ifndef GEOS370
+		setError("GEOS >= 3.7 needed for CCW");
+		return;
+	#else
+	size_t n = size();
+	if (n == 0) return;
+	if (geoms[0].gtype != polygons) return;
+	for (size_t i=0; i<n; i++) {
+		for (size_t j=0; j<geoms[i].parts.size(); j++) {
+			if (!geoms[i].parts[j].is_CCW()) {
+				std::reverse(geoms[i].parts[j].x.begin(), geoms[i].parts[j].x.end());
+				std::reverse(geoms[i].parts[j].y.begin(), geoms[i].parts[j].y.end());
+				for (size_t k=0; k<geoms[i].parts[j].nHoles(); k++) {
+					std::reverse(geoms[i].parts[j].holes[k].x.begin(), geoms[i].parts[j].holes[k].x.end());
+					std::reverse(geoms[i].parts[j].holes[k].y.begin(), geoms[i].parts[j].holes[k].y.end());
+				}
+			}
+		}
+	}
+	#endif
+}
+
+
