@@ -236,12 +236,39 @@ void SpatRaster::readValues(std::vector<double> &out, size_t row, size_t nrows,
 		return;
 	}
 
+	out.resize(0);
 	if (!hasValues()) {
 		out.resize(nrows * ncols * nlyr(), NAN);
 		addWarning("raster has no values");
 		return; // or NAs?
 	}
 
+	unsigned n = nsrc();
+	out.reserve(nrows * ncols * nlyr());
+	for (size_t src=0; src<n; src++) {
+		if (source[src].memory) {
+			readChunkMEM(out, src, row, nrows, col, ncols);
+		} else {
+			// read from file
+			#ifdef useGDAL
+			readChunkGDAL(out, src, row, nrows, col, ncols);
+			#endif // useGDAL
+		}
+	}
+	return;
+}
+
+
+void SpatRaster::readValuesWhileWriting(std::vector<double> &out, size_t row, size_t nrows, size_t col, size_t ncols){
+
+	if (((row + nrows) > nrow()) || ((col + ncols) > ncol())) {
+		setError("invalid rows/columns");
+		return;
+	}
+
+	if ((nrows==0) | (ncols==0)) {
+		return;
+	}
 
 	unsigned n = nsrc();
 	out.resize(0);
@@ -286,11 +313,14 @@ bool SpatRaster::readAll() {
 		}
 	}
 	readStop();
-	if (n>1) source.resize(1);
+	if (n > 1) source.resize(1);
+
+	source[0].hasWindow = false;
 	return true;
 }
 
 
+
 std::vector<double> SpatRaster::getValues(long lyr, SpatOptions &opt) {
 
 	std::vector<double> out;
