@@ -1,6 +1,18 @@
 
+#ifndef SORT_GUARD
+#define SORT_GUARD
+
+#include <numeric>
+
+std::vector<std::size_t> sort_order_nan_a(const std::vector<double> &x);
+std::vector<std::size_t> sort_order_nan_d(const std::vector<double> &x);
+std::vector<std::size_t> sort_order_nal_a(const std::vector<long> &x);
+std::vector<std::size_t> sort_order_nal_d(const std::vector<long> &x);
+std::vector<std::size_t> sort_order_nas_a(const std::vector<std::string> &x);
+std::vector<std::size_t> sort_order_nas_d(const std::vector<std::string> &x);
+
 template <typename T>
-std::vector<std::size_t> sort_order_d(const std::vector<T> &x){
+std::vector<std::size_t> sort_order_a(const std::vector<T> &x){
 	std::vector<std::size_t> p(x.size());
 	std::iota(p.begin(), p.end(), 0);
 	std::sort(p.begin(), p.end(),
@@ -9,7 +21,7 @@ std::vector<std::size_t> sort_order_d(const std::vector<T> &x){
 }
 
 template <typename T>
-std::vector<std::size_t> sort_order_a(const std::vector<T> &x){
+std::vector<std::size_t> sort_order_d(const std::vector<T> &x){
 	std::vector<std::size_t> p(x.size());
 	std::iota(p.begin(), p.end(), 0);
 	std::sort(p.begin(), p.end(),
@@ -17,6 +29,7 @@ std::vector<std::size_t> sort_order_a(const std::vector<T> &x){
 	return p;
 }
 
+
 template <typename T>
 void permute(std::vector<T> &x, const std::vector<std::size_t> &p) {
 	std::vector<bool> done(x.size());
@@ -36,3 +49,4 @@ void permute(std::vector<T> &x, const std::vector<std::size_t> &p) {
 	}
 }
 
+#endif
