@@ -21,6 +21,7 @@
 #include <string>
 #include "NA.h"
 #include "string_utils.h"
+#include "sort.h"
 
 
 SpatDataFrame::SpatDataFrame() {}
@@ -112,7 +113,7 @@ SpatDataFrame SpatDataFrame::subset_rows(std::vector<unsigned> range) {
 
 	SpatDataFrame out;
 	unsigned nr = nrow();
-	if (range.size() > 0) {
+	if (!range.empty()) {
 		for (int i = range.size(); i>0; i--) {
 			if (range[i-1] > nr) {
 				range.erase(range.begin() + i-1);
@@ -233,7 +234,7 @@ unsigned SpatDataFrame::ncol() {
 
 unsigned SpatDataFrame::nrow() {
 	unsigned n;
-	if (itype.size() == 0) {
+	if (itype.empty()) {
 		n = 0;
 	} else {
 		if (itype[0] == 0) {
@@ -260,9 +261,8 @@ void SpatDataFrame::add_row() {
 	for (size_t i=0; i < dv.size(); i++) {
 		dv[i].push_back(NAN);
 	}
-	long longNA = NA<long>::value;
 	for (size_t i=0; i < iv.size(); i++) {
-		iv[i].push_back(longNA);
+		iv[i].push_back(NAL);
 	}
 	for (size_t i=0; i < sv.size(); i++) {
 		sv[i].push_back(NAS);
@@ -270,9 +270,8 @@ void SpatDataFrame::add_row() {
 	for (size_t i=0; i < bv.size(); i++) {
 		bv[i].push_back(2);
 	}
-	SpatTime_t timeNA = NA<SpatTime_t>::value;
 	for (size_t i=0; i < tv.size(); i++) {
-		tv[i].push_back(timeNA);
+		tv[i].push_back(NAT);
 	}
 	for (size_t i=0; i < fv.size(); i++) {
 		fv[i].push_back(0);
@@ -350,7 +349,7 @@ void SpatDataFrame::resize_rows(unsigned n) {
 }
 
 void SpatDataFrame::remove_rows(std::vector<unsigned> r) {
-	if (r.size() == 0) return;
+	if (r.empty()) return;
 	//sort(r.begin(), r.end(), std::greater<unsigned>());
 	sort(r.begin(), r.end());
 	r.erase(std::unique(r.begin(), r.end()), r.end());
@@ -587,7 +586,7 @@ bool SpatDataFrame::cbind(SpatDataFrame &x) {
 		} else if (x.itype[i] == 3) {
 			std::vector<int8_t> d = x.getB(i);
 			if (!add_column(d, nms[i])) return false;
-		} else if (x.itype[i] == 3) {
+		} else if (x.itype[i] == 4) {
 			SpatTime_v d = x.getT(i);
 			if (!add_column(d, nms[i])) return false;
 		} else {
@@ -598,68 +597,101 @@ bool SpatDataFrame::cbind(SpatDataFrame &x) {
 	return true;
 }
 
-
 bool SpatDataFrame::rbind(SpatDataFrame &x) {
 
 	size_t nr1 = nrow();
 	size_t nr2 = x.nrow();
-//	size_t nc1 = ncol();
+	size_t nc1 = ncol();
 	size_t nc2 = x.ncol();
+	std::vector<std::string> nms = names;
+	std::vector<std::string> xnms = x.names;
+
+	if (nc2 == nc1) {
+		bool same = true;
+		for (size_t i=0; i<nc2; i++) {
+			if ((nms[i] != xnms[i]) || (itype[i] != x.itype[i])) {
+				same = false;
+				break;
+			}
+		}
+		// the simplest case; no issue with duplicate names
+		if (same) {
+			for (size_t i=0; i<nc2; i++) {
+				size_t j = iplace[i];
+				if (itype[i] == 0) {
+					dv[j].insert(dv[j].end(), x.dv[j].begin(), x.dv[j].end());
+				} else if (itype[i] == 1) {
+					iv[j].insert(iv[j].end(), x.iv[j].begin(), x.iv[j].end());
+				} else if (itype[i] == 2) {
+					sv[j].insert(sv[j].end(), x.sv[j].begin(), x.sv[j].end());
+				} else if (itype[i] == 3) {
+					bv[j].insert(bv[j].end(), x.bv[j].begin(), x.bv[j].end());
+				} else if (itype[i] == 4) {
+					tv[j].x.insert(tv[j].x.end(), x.tv[j].x.begin(), x.tv[j].x.end());
+				} else {
+					//should check the levels, needs proper method
+					fv[j].v.insert(fv[j].v.end(), x.fv[j].v.begin(), x.fv[j].v.end());
+				}			
+			}
+		}
+	}
+
+	make_unique_names(nms);
+	set_names(nms);
+	make_unique_names(xnms);
+	x.set_names(xnms);
+
+
+// should check for duplicate names and warn
 
 //first add new columns
-	std::vector<std::string> nms = names;
 	for (size_t i=0; i<nc2; i++) {
-		int j = where_in_vector(x.names[i], nms, false);
+		int j = where_in_vector(xnms[i], nms, false);
 		if (j < 0) { // not in df
 			size_t b = x.iplace[i];
-			add_column(x.itype[i], x.names[i]);
+			add_column(x.itype[i], xnms[i]);
 			if (x.itype[i] == 0) {
 				size_t a = dv.size()-1;
-				dv[a].insert(dv[a].begin()+nr1,
-					x.dv[b].begin(), x.dv[b].end());
+				dv[a].resize(nr1, NAN);
+				dv[a].insert(dv[a].end(), x.dv[b].begin(), x.dv[b].end());
 			} else if (x.itype[i] == 1) {
 				size_t a = iv.size()-1;
-				iv[a].insert(iv[a].begin()+nr1,
-					x.iv[b].begin(), x.iv[b].end());
+				iv[a].resize(nr1, NAL);
+				iv[a].insert(iv[a].end(), x.iv[b].begin(), x.iv[b].end());
 			} else if (x.itype[i] == 2) {
 				size_t a = sv.size()-1;
-				sv[a].insert(sv[a].begin()+nr1,
-					x.sv[b].begin(), x.sv[b].end());
+				sv[a].resize(nr1, NAS);
+				sv[a].insert(sv[a].end(), x.sv[b].begin(), x.sv[b].end());
 			} else if (x.itype[i] == 3) {
 				size_t a = bv.size()-1;
-				bv[a].insert(bv[a].begin()+nr1,
-					x.bv[b].begin(), x.bv[b].end());
+				bv[a].resize(nr1, 2);
+				bv[a].insert(bv[a].end(), x.bv[b].begin(), x.bv[b].end());
 			} else if (x.itype[i] == 4) {
 				size_t a = tv.size()-1;
-				tv[a].x.insert(tv[a].x.begin()+nr1,
-					x.tv[b].x.begin(), x.tv[b].x.end());
+				tv[a].resize(nr1, NAT);
+				tv[a].x.insert(tv[a].x.end(), x.tv[b].x.begin(), x.tv[b].x.end());
 			} else {
 				size_t a = fv.size()-1;
-				fv[a].v.insert(fv[a].v.begin()+nr1,
-					x.fv[b].v.begin(), x.fv[b].v.end());
+				fv[a].resize(nr1, 0);
+				fv[a].v.insert(fv[a].v.end(), x.fv[b].v.begin(), x.fv[b].v.end());
 			}
 		} else {
 			size_t a = iplace[j];
 			size_t b = x.iplace[i];
 			if (itype[j] == x.itype[i]) {
 				if (itype[j] == 0) {
-					dv[a].insert(dv[a].begin()+nr1,
-					x.dv[b].begin(), x.dv[b].end());
+					dv[a].insert(dv[a].end(), x.dv[b].begin(), x.dv[b].end());
 				} else if (itype[j] == 1) {
-					iv[a].insert(iv[a].begin()+nr1,
-					x.iv[b].begin(), x.iv[b].end());
+					iv[a].insert(iv[a].end(), x.iv[b].begin(), x.iv[b].end());
 				} else if (itype[j] == 2) {
-					sv[a].insert(sv[a].begin()+nr1,
-					x.sv[b].begin(), x.sv[b].end());
+					sv[a].insert(sv[a].end(), x.sv[b].begin(), x.sv[b].end());
 				} else if (itype[j] == 3) {
-					bv[a].insert(bv[a].begin()+nr1,
-					x.bv[b].begin(), x.bv[b].end());
+					bv[a].insert(bv[a].end(), x.bv[b].begin(), x.bv[b].end());
 				} else if (itype[j] == 4) {
-					tv[a].x.insert(tv[a].x.begin()+nr1,
-					x.tv[b].x.begin(), x.tv[b].x.end());
+					tv[a].x.insert(tv[a].x.end(), x.tv[b].x.begin(), x.tv[b].x.end());
 				} else {
-					fv[a].v.insert(fv[a].v.begin()+nr1,
-					x.fv[b].v.begin(), x.fv[b].v.end());
+					//should check the levels, needs proper method
+					fv[a].v.insert(fv[a].v.end(), x.fv[b].v.begin(), x.fv[b].v.end());
 				}
 			} else {
 				if (itype[j] == 2) {
@@ -676,7 +708,7 @@ bool SpatDataFrame::rbind(SpatDataFrame &x) {
 						long longNA = NA<long>::value;
 						for (size_t k=0; k<nr2; k++) {
 							if (x.iv[b][k] == longNA) {
-								sv[a].push_back("____NA_+");
+								sv[a].push_back(NAS);
 							} else {
 								sv[a].push_back(std::to_string(x.iv[b][k]));
 							}
@@ -1165,3 +1197,76 @@ size_t SpatDataFrame::strwidth(unsigned i) {
 	return m;
 }
 
+
+
+SpatDataFrame SpatDataFrame::sortby(std::string field, bool descending) {
+	SpatDataFrame out = *this;
+
+	std::vector<std::string> nms = get_names();
+	int i = where_in_vector(field, nms, false);
+	if (i < 0) { // not in df
+		out.setError("unknown variable: " + field);
+		return out;
+	}
+	size_t j = iplace[i];
+	std::vector<std::size_t> order;
+	if (itype[i] == 0) {
+		if (descending) {
+			order = sort_order_nan_d(dv[j]);
+		} else {
+			order = sort_order_nan_a(dv[j]);			
+		}
+	} else if (itype[i] == 1) {
+		if (descending) {
+			order = sort_order_nal_d(iv[j]);
+		} else {
+			order = sort_order_nal_a(iv[j]);			
+		}
+	} else if (itype[i] == 2) {
+		if (descending) {
+			order = sort_order_nas_d(sv[j]);
+		} else {
+			order = sort_order_nas_a(sv[j]);			
+		}
+	} else if (itype[i] == 3) {
+		if (descending) {
+			order = sort_order_d(bv[j]);
+		} else {
+			order = sort_order_a(bv[j]);			
+		}
+	} else if (itype[i] == 4) {
+		if (descending) {
+			order = sort_order_d(tv[j].x);
+		} else {
+			order = sort_order_a(tv[j].x);			
+		}
+	} else {
+		if (descending) {
+			order = sort_order_d(fv[j].v);
+		} else {
+			order = sort_order_a(fv[j].v);			
+		}
+	}
+
+	for (size_t i=0; i<dv.size(); i++) {
+		permute(out.dv[i], order);
+	}
+	for (size_t i=0; i<iv.size(); i++) {
+		permute(out.iv[i], order);
+	}
+	for (size_t i=0; i<sv.size(); i++) {
+		permute(out.sv[i], order);
+	}
+	for (size_t i=0; i<bv.size(); i++) {
+		permute(out.bv[i], order);
+	}
+	for (size_t i=0; i<tv.size(); i++) {
+		permute(out.tv[i].x, order);
+	}
+	for (size_t i=0; i<fv.size(); i++) {
+		permute(out.fv[i].v, order);
+	}
+	
+	return out;
+}
+
