@@ -25,6 +25,8 @@
 #include "spatBase.h"
 #include "spatTime.h"
 #include "spatFactor.h"
+#include <cstdint>
+#include <limits>
 
 class SpatDataFrame {
 	public:
@@ -51,6 +53,8 @@ class SpatDataFrame {
 		std::vector<SpatTime_v> tv;
 		std::vector<SpatFactor> fv;		
 		std::string NAS = "____NA_+";
+		long NAL = std::numeric_limits<long>::min();
+		SpatTime_t NAT = std::numeric_limits<SpatTime_t>::min();
 		
 		unsigned nrow();
 		unsigned ncol();
@@ -134,6 +138,8 @@ class SpatDataFrame {
 		std::vector<std::string> one_string();
 		SpatDataFrame unique();
 		size_t strwidth(unsigned i);
+
+		SpatDataFrame sortby(std::string field, bool descending);
 };
 
 #endif //SPATDATAFRAME_GUARD
