@@ -36,17 +36,17 @@ SpatRaster::SpatRaster(std::string fname, std::vector<int> subds, std::vector<st
 }
 
 
-SpatRaster::SpatRaster(std::vector<std::string> fname, std::vector<int> subds, std::vector<std::string> subdsname, bool multi, std::vector<std::string> drivers, std::vector<std::string> options, std::vector<size_t> x) {
+SpatRaster::SpatRaster(std::vector<std::string> fname, std::vector<int> subds, std::vector<std::string> subdsname, bool multi, std::vector<std::string> drivers, std::vector<std::string> options,std::vector<size_t> xyz) {
 // argument "x" is ignored. It is only there to have four arguments such that the  module
 // can distinguish this constructor from another with three arguments.
-	if (fname.size() == 0) {
+	if (fname.empty()) {
 		setError("no filename");
 		return;
 	}
 
 #ifdef useGDAL
 	if (multi) {
-		constructFromFileMulti(fname[0], subds, subdsname, drivers, options, x);
+		constructFromFileMulti(fname[0], subds, subdsname, drivers, options, xyz);
 		return;
 	}
 
@@ -126,6 +126,17 @@ SpatRaster SpatRaster::dropSource() {
 }
 */
 
+
+SpatRaster SpatRaster::subsetSource(size_t snr) {
+	if (snr >= source.size()) {
+		SpatRaster out;
+		out.setError("invalid source number");
+		return out;
+	}
+	SpatRaster out(source[snr]);
+	return out;
+}
+
 bool SpatRaster::hasValues() {
 //	if (source.size() == 0) {
 //		return false;
@@ -138,6 +149,7 @@ bool SpatRaster::hasValues() {
 SpatRaster::SpatRaster(std::vector<unsigned> rcl, std::vector<double> ext, std::string crs) {
 
 	SpatRasterSource s;
+	rcl.resize(3, 1);
 	s.nrow=rcl[0];
 	s.ncol=rcl[1];
 	s.extent.xmin = ext[0];
@@ -163,7 +175,7 @@ SpatRaster::SpatRaster(std::vector<unsigned> rcl, std::vector<double> ext, std::
 	if (!s.srs.set( crs, msg )) {
 		setError(msg);
 		return;
-	} else if (msg != "") {
+	} else if (!msg.empty()) {
 		addWarning(msg);
 	}
 
@@ -203,7 +215,7 @@ SpatRaster::SpatRaster(unsigned nr, unsigned nc, unsigned nl, SpatExtent ext, st
 	if (!s.srs.set(crs, msg )) {
 		setError(msg);
 		return;
-	} else if (msg != "") {
+	} else if (!msg.empty()) {
 		addWarning(msg);
 	}
 #else
@@ -236,7 +248,7 @@ SpatRaster::SpatRaster(const SpatRaster &r) {
 
 
 
-SpatRaster SpatRaster::geometry(long nlyrs, bool properties, bool time, bool units) {
+SpatRaster SpatRaster::geometry(long nlyrs, bool properties, bool time, bool units, bool keeptags) {
 	SpatRasterSource s;
 	//s.values.resize(0);
 	s.nrow = nrow();
@@ -249,6 +261,7 @@ SpatRaster SpatRaster::geometry(long nlyrs, bool properties, bool time, bool uni
 	long nl = nlyr();
 	bool keepnlyr = ((nlyrs == nl) || (nlyrs < 1));
 	nlyrs = (keepnlyr) ? nlyr(): nlyrs;
+	
 	if (properties) {
 		s.hasColors = hasColors();
 		s.cols = getColors();
@@ -269,6 +282,19 @@ SpatRaster SpatRaster::geometry(long nlyrs, bool properties, bool time, bool uni
 			s.hasUnit = true;
 			s.unit = getUnit();
 		}
+		
+		std::vector<std::string> un = getSourceNames();
+		std::sort(un.begin(), un.end() );
+		un.erase(std::unique(un.begin(), un.end()), un.end());
+		if (un.size() == 1) {
+			s.source_name = un[0];
+		}
+		un = getLongSourceNames();
+		std::sort(un.begin(), un.end() );
+		un.erase(std::unique(un.begin(), un.end()), un.end());
+		if (un.size() == 1) {
+			s.source_name_long = un[0];
+		}
 	} else {
 		for (size_t i=0; i < s.nlyr; i++) {
 			nms.push_back("lyr" + std::to_string(i+1));
@@ -281,20 +307,26 @@ SpatRaster SpatRaster::geometry(long nlyrs, bool properties, bool time, bool uni
 		out.rgbtype = rgbtype;
 		out.rgblyrs = rgblyrs;
 	}	
+	if (keeptags) {
+		out.tags = tags;
+		out.lyrTags = lyrTags;
+	}
 	return out;
 }
 
 
-SpatRaster SpatRaster::geometry_opt(long nlyrs, bool properties, bool time, bool units, bool datatype, SpatOptions &opt) {
+SpatRaster SpatRaster::geometry_opt(long nlyrs, bool properties, bool time, bool units, bool tags, bool datatype, SpatOptions &opt) {
 
 	if (datatype && hasValues() && (!opt.datatype_set)) {
 		std::vector<std::string> dt = getDataType(true);
-		if ((dt.size() == 1) && (dt[0] != "")) {
-			opt.set_datatype(dt[0]);
+		if ((dt.size() == 1) && !dt[0].empty()) {
+			if (!hasScaleOffset()) {
+				opt.set_datatype(dt[0]);
+			}
 		}
 	}	
 	
-	return geometry(nlyrs, properties, time, units);
+	return geometry(nlyrs, properties, time, units, tags);
 }
 
 SpatRaster SpatRaster::deepCopy() {
@@ -333,18 +365,18 @@ SpatRaster SpatRaster::setResolution(double xres, double yres) {
 
 
 size_t SpatRaster::ncol() {
-	if (source.size() > 0) {
-		return source[0].ncol;
-	} else {
+	if (source.empty()) {
 		return 0;
+	} else {
+		return source[0].ncol;
 	}
 }
 
 size_t SpatRaster::nrow() {
-	if (source.size() > 0) {
-		return source[0].nrow;
+	if (source.empty()) {
+		return source[0].ncol;
 	} else {
-		return 0;
+		return source[0].nrow;
 	}
 }
 
@@ -420,7 +452,14 @@ std::vector<double> SpatRaster::range_max() {
 
 
 bool SpatRaster::is_lonlat() {
-	return source[0].srs.is_lonlat();
+	if (source[0].srs.is_lonlat()) {
+		SpatExtent e = getExtent();
+		if ((e.xmin < -181) || (e.xmax > 361) || (e.ymin < -90.001) || (e.ymax > 90.001)) {
+			addWarning("coordinates are out of range for lon/lat");
+		}
+		return true;
+	}
+	return false;
 }
 
 bool SpatRaster::could_be_lonlat() {
@@ -507,7 +546,7 @@ bool SpatRaster::setSRS(std::string crs) {
 	if (!srs.set(crs, msg )) {
 		addWarning("Cannot set raster SRS: "+ msg);
 		return false;
-	} else if (msg != "") {
+	} else if (!msg.empty()) {
 		addWarning(msg);
 	}
 
@@ -749,7 +788,7 @@ std::string SpatRaster::getTimeZone() {
 
 bool SpatRaster::setTime(std::vector<int_64> time, std::string step, std::string zone) {
 
-	if (time.size() == 0 || step == "remove") {
+	if (time.empty() || step == "remove") {
 		for (size_t i=0; i<source.size(); i++)	{
 			source[i].time = std::vector<int_64> (source[i].nlyr);
 			source[i].timestep = "";
@@ -798,7 +837,7 @@ std::vector<double> SpatRaster::getDepth() {
 
 bool SpatRaster::setDepth(std::vector<double> depths) {
 
-	if (depths.size() == 0) {
+	if (depths.empty()) {
 		for (size_t i=0; i<source.size(); i++)	{
 			source[i].depth = std::vector<double>(source[i].nlyr);
 		}
@@ -828,7 +867,7 @@ bool SpatRaster::setDepth(std::vector<double> depths) {
 bool SpatRaster::setUnit(std::vector<std::string> units) {
 	if (units.size() == 1) {
 		bool hu = true;
-		if (units[0] == "") {
+		if (units[0].empty()) {
 			hu = false;
 		}
         for (size_t i=0; i<source.size(); i++)	{
@@ -886,13 +925,24 @@ double SpatRaster::yres() {
 }
 
 
+std::vector<bool> SpatRaster::is_rotated() {
+	std::vector<bool> b(source.size(), false);
+	for (size_t i=0; i<source.size(); i++) {
+		if (source[i].rotated) {
+			b[i] = true;
+		}
+	}
+	return b;
+}
+
+
 bool SpatRaster::valid_sources(bool files, bool rotated) {
 	std::vector<std::string> ff;
 	for (size_t i=0; i<source.size(); i++) {
 		std::string f = source[i].filename;
-		if (f == "") continue;
+		if (f.empty()) continue;
 		if (files) {
-			std::size_t found = f.find(":"); // perhaps http: or PG:xxx
+			std::size_t found = f.find(':'); // perhaps http: or PG:xxx
 			if ((found == 1) || (found == std::string::npos)) {
 				if (!file_exists(f)) {
 					setError("missing source: " + f);
@@ -1052,45 +1102,41 @@ SpatRaster SpatRaster::replace(SpatRaster x, unsigned layer, SpatOptions &opt) {
 }
 
 
-SpatRaster SpatRaster::makeCategorical(unsigned layer, SpatOptions &opt) {
+SpatRaster SpatRaster::makeCategorical(long layer, SpatOptions &opt) {
 
+	SpatRaster out;
 	if (!hasValues()) {
-		SpatRaster out;
 		out.setError("cannot make categories if the raster has no values");
 		return out;
 	}
 
-	std::vector<unsigned> lyrs = {layer};
+	SpatRaster r;
 	SpatOptions fopt(opt);
-	SpatRaster r = subset(lyrs, fopt);
-
+	if (layer >= 0) {
+		if (layer > (long) nlyr()) {
+			out.setError("layer number is too high");
+			return out;
+		}
+		std::vector<unsigned> lyrs = {(unsigned) layer};
+		r = subset(lyrs, fopt);
+	} else {
+		r = *this;
+	}	
 	r.math2("round", 0, fopt);
-
-	std::vector<std::vector<double>> u = r.unique(false, true, fopt);
-/*
-	std::vector<double> id(u[0].size());
-	std::iota(id.begin(), id.end(), 0);
-	std::vector<std::vector<double>> rcl(2);
-	rcl[0] = u[0];
-	rcl[1] = id;
-	r = r.reclassify(rcl, true, true, true, false, false, fopt);
-
-	std::vector<std::string> s(id.size());
-	for (size_t i=0; i<s.size(); i++) {
-		s[i] = std::to_string((int)u[0][i]);
-	}
-*/
-
-	std::vector<long> uu;
-	std::vector<std::string> s(u[0].size());
-	for (size_t i=0; i<s.size(); i++) {
-		uu[i] = (long)u[0][i];
-		s[i] = std::to_string(uu[i]);
-	}
+	std::vector<std::vector<double>> u = r.unique(true, NAN, true, fopt);
 	std::vector<std::string> names = r.getNames();
-	r.setLabels(0, uu, s, names[0]);
-
-	if (nlyr() == 1) {
+	
+	for (size_t i=0; i<r.nlyr(); i++) { 
+		std::vector<long> uu(u[i].size());
+		std::vector<std::string> s(u[i].size());
+		for (size_t j=0; j<s.size(); j++) {
+			uu[j] = (long)u[i][j];
+			s[j] = std::to_string(uu[j]);
+		}
+		r.setLabels(i, uu, s, names[i]);
+	}
+	
+	if (nlyr() == r.nlyr()) {
 		return r;
 	} else {
 		return replace(r, layer, opt);
@@ -1106,7 +1152,7 @@ bool SpatRaster::createCategories(unsigned layer, SpatOptions &opt) {
 	}
 	std::vector<unsigned> lyrs(1, layer);
 	SpatRaster r = subset(lyrs, opt);
-	std::vector<std::vector<double>> u = r.unique(false, true, opt);
+	std::vector<std::vector<double>> u = r.unique(false, NAN, true, opt);
     std::vector<unsigned> sl = findLyr(layer);
 
 	std::vector<std::string> s(u[0].size());
@@ -1162,6 +1208,29 @@ std::vector<std::string> SpatRaster::dataType() {
 }
 
 
+std::vector<std::vector<std::string>> SpatRaster::getMetadata(bool layers) {
+	std::vector<std::vector<std::string>> d;
+	size_t n = nsrc();
+	if (layers) {
+		d.reserve(nlyr());
+		for (size_t i=0; i<n; i++) {
+			if (source[i].bmdata.empty()) {
+				d.resize(d.size() + source[i].nlyr);
+			} else {
+				d.insert(d.end(), source[i].bmdata.begin(), source[i].bmdata.end());
+			}
+		}
+	} else {
+		d.resize(n);
+		for (size_t i=0; i<n; i++) {
+			if (!source[i].smdata.empty()) {
+				d[i] = source[i].smdata;
+			}
+		}
+	}
+	return d;
+}
+
 
 
 bool SpatRaster::setLabels(unsigned layer, std::vector<long> values, std::vector<std::string> labels, std::string name) {
@@ -1174,7 +1243,7 @@ bool SpatRaster::setLabels(unsigned layer, std::vector<long> values, std::vector
 		setError("size of values is not the same as the size of labels");
 		return(false);
 	}
-	if (values.size() == 0) {
+	if (values.empty()) {
 		addWarning("no labels");
 		return(true);
 	}
@@ -1182,7 +1251,7 @@ bool SpatRaster::setLabels(unsigned layer, std::vector<long> values, std::vector
     std::vector<unsigned> sl = findLyr(layer);
 
 	SpatCategories cats;
-	cats.d.add_column(values, "value");
+	cats.d.add_column(values, "ID");
 	cats.d.add_column(labels, name);
 	cats.index = 1;
 
@@ -1267,6 +1336,15 @@ std::vector<std::vector<double>> SpatRaster::getScaleOffset() {
 	return so;
 }
 
+bool SpatRaster::hasScaleOffset() {
+	for (size_t i=0; i<source.size(); i++) {
+		for (size_t j=0; j<source[i].has_scale_offset.size(); j++) {
+			if (source[i].has_scale_offset[j]) return true;
+		}
+	}
+	return false;
+}
+
 bool SpatRaster::setScaleOffset(std::vector<double> sc, std::vector<double> of) {
 	size_t n = sc.size();
 	size_t nl = nlyr();
@@ -1333,23 +1411,25 @@ std::vector<std::string> SpatRaster::getLabels(unsigned layer) {
 	std::vector<SpatCategories> cats = getCategories();
 	SpatCategories cat = cats[layer];
 
-	unsigned nc = cat.d.ncol();
-	if (nc == 0) return out;
+	int nc = cat.d.ncol();
+	if (nc <= 0) return out;
 
 	cat.index = cat.index > (nc-1) ? (nc-1) : cat.index;
 	out = cat.d.as_string(cat.index);
 	return out;
 }
 
-bool SpatRaster::setCatIndex(unsigned layer, unsigned index) {
+bool SpatRaster::setCatIndex(unsigned layer, int index) {
 	if (layer > (nlyr()-1)) {
 		return(false);
 	}
     std::vector<unsigned> sl = findLyr(layer);
-	unsigned nc = source[sl[0]].cats[sl[1]].d.ncol();
+	int nc = source[sl[0]].cats[sl[1]].d.ncol();
 	if (index < nc) {
-		source[sl[0]].cats[sl[1]].index  = index;
-		source[sl[0]].names[sl[1]] = source[sl[0]].cats[sl[1]].d.names[index];
+		source[sl[0]].cats[sl[1]].index = index;
+		if (index >= 0) {
+			source[sl[0]].names[sl[1]] = source[sl[0]].cats[sl[1]].d.names[index];
+		}
 		return true;
 	} else {
 		return false;
@@ -1358,7 +1438,7 @@ bool SpatRaster::setCatIndex(unsigned layer, unsigned index) {
 
 int SpatRaster::getCatIndex(unsigned layer) {
 	if (layer > (nlyr()-1)) {
-		return(-1);
+		return( -1 );
 	}
     std::vector<unsigned> sl = findLyr(layer);
 	return source[sl[0]].cats[sl[1]].index;
@@ -1378,7 +1458,7 @@ SpatRaster SpatRaster::dropLevels() {
 	std::vector<SpatCategories> cats = getCategories();
 	SpatOptions opt;
 	SpatRaster out = *this;
-	std::vector<std::vector<double>> uvv = unique(true, true, opt);
+	std::vector<std::vector<double>> uvv = unique(true, NAN, true, opt);
 	for (size_t i=0; i<hascats.size(); i++) {
 		if (hascats[i]) {
 			SpatCategories lyrcats = cats[i];
@@ -1423,9 +1503,11 @@ std::vector<SpatDataFrame> SpatRaster::getColors() {
 
 bool SpatRaster::setColors(size_t layer, SpatDataFrame cols) {
 	if (cols.ncol() < 4 || cols.ncol() > 5) {
+		setError("n columns should be 4 or 5");
 		return false;
 	}
 	if (layer >= nlyr()) {
+		setError("layer > nlyr");
 		return false;
 	}
 	if (cols.ncol() == 4) {
@@ -1476,7 +1558,7 @@ std::vector<bool> SpatRaster::hasColors() {
 }
 
 
-std::vector<double> SpatRaster::cellFromXY (std::vector<double> x, std::vector<double> y) {
+std::vector<double> SpatRaster::cellFromXY (std::vector<double> x, std::vector<double> y, double missing) {
 // size of x and y should be the same
 
 	size_t size = x.size();
@@ -1503,7 +1585,7 @@ std::vector<double> SpatRaster::cellFromXY (std::vector<double> x, std::vector<d
 		long nr = nrow();
 		long nc = ncol();
 		if (row < 0 || row >= nr || col < 0 || col >= nc) {
-			cells[i] = NAN;
+			cells[i] = missing;
 		} else {
 			cells[i] = row * ncol() + col;
 		}
@@ -1513,10 +1595,10 @@ std::vector<double> SpatRaster::cellFromXY (std::vector<double> x, std::vector<d
 }
 
 
-double SpatRaster::cellFromXY (double x, double y) {
+double SpatRaster::cellFromXY (double x, double y, double missing) {
 	std::vector<double> X = {x};
 	std::vector<double> Y = {y};
-	std::vector<double> cell = cellFromXY(X, Y);
+	std::vector<double> cell = cellFromXY(X, Y, missing);
 	return  cell[0];
 }
 
@@ -1660,6 +1742,27 @@ int_64 SpatRaster::rowFromY(double y) {
 }
 
 
+void SpatRaster::xyFromCell( std::vector<std::vector<double>> &xy ) {
+	
+	SpatExtent extent = getExtent();
+	double xmin = extent.xmin;
+	double ymax = extent.ymax;
+	double yr = yres();
+	double xr = xres();
+	size_t nr = nrow();
+	size_t nc = ncol();
+
+	xy[0].reserve(ncell()+2); 
+	xy[1].reserve(ncell()+2); 
+	for (size_t i = 0; i<nr; i++) {
+		for (size_t j = 0; j<nc; j++) {
+			xy[0].push_back( xmin + (j + 0.5) * xr );
+			xy[1].push_back( ymax - (i + 0.5) * yr );
+		}
+	}
+}
+
+
 std::vector<std::vector<double>> SpatRaster::xyFromCell( std::vector<double> &cell) {
 	size_t n = cell.size();
 	SpatExtent extent = getExtent();
@@ -2093,6 +2196,63 @@ std::vector<std::vector<double>> SpatRaster::as_points_value(const double& targe
 	return xyFromCell(cells);
 }
 
+
+
+std::vector<std::vector<double>> SpatRaster::coordinates(bool narm, bool nall, SpatOptions &opt) {
+
+    std::vector<std::vector<double>> xy(2);
+
+	if ( !(narm) || (!hasValues()) ) {
+        xyFromCell(xy);
+		return xy;
+	}
+
+	BlockSize bs = getBlockSize(opt);
+
+	if (!readStart()) {
+		return(xy);
+	}
+	size_t nc = ncol();
+	unsigned nl = nlyr();
+	std::vector<double> v;
+	for (size_t i = 0; i < bs.n; i++) {
+		readValues(v, bs.row[i], bs.nrows[i], 0, nc);
+        size_t off1 = (bs.row[i] * nc);
+ 		size_t vnc = bs.nrows[i] * nc;
+		for (size_t j=0; j<vnc; j++) {
+			if (nall) {
+				bool allna = true;
+				size_t off2 = 0;
+				for (size_t lyr=0; lyr<nl; lyr++) {
+					if (!std::isnan(v[off2+j])) {
+						allna = false;
+						continue;
+					}
+					off2 += vnc;
+				}
+				if (allna) continue;
+			} else {
+				bool foundna = false;
+				size_t off2 = 0;
+				for (size_t lyr=0; lyr<nl; lyr++) {
+					if (std::isnan(v[off2+j])) {
+						foundna = true;
+						continue;
+					}
+					off2 += vnc;
+				}
+				if (foundna) continue;
+			}
+			std::vector<std::vector<double>> xyc = xyFromCell( off1+j );
+			xy[0].push_back(xyc[0][0]);
+			xy[1].push_back(xyc[1][0]);
+		}
+	}
+	readStop();
+	return(xy);
+}
+
+
 std::vector<std::vector<double>> SpatRaster::cells_notna(SpatOptions &opt) {
 
 	std::vector<std::vector<double>> out(2);
@@ -2179,65 +2339,8 @@ void getCorners(std::vector<double> &x,  std::vector<double> &y, const double &X
 	y[4] = y[0];
 }
 
-/*
-SpatVector SpatRaster::as_polygons(bool values, bool narm) {
-	if (!values) narm=false;
-	SpatVector v;
-	SpatGeom g;
-	g.gtype = polygons;
-	double xr = xres()/2;
-	double yr = yres()/2;
-	std::vector<double> x(5);
-	std::vector<double> y(5);
-	if (!values) {
-		std::vector<double> cells(ncell()) ;
-		std::iota (std::begin(cells), std::end(cells), 0);
-		std::vector< std::vector<double> > xy = xyFromCell(cells);
-		for (size_t i=0; i<ncell(); i++) {
-			getCorners(x, y, xy[0][i], xy[1][i], xr, yr);
-			SpatPart p(x, y);
-			g.addPart(p);
-			v.addGeom(g);
-			g.parts.resize(0);
-		}
-	} else {
-		SpatRaster out = geometry();
-		unsigned nl = nlyr();
-		std::vector<std::vector<double> > att(ncell(), std::vector<double> (nl));
 
-		BlockSize bs = getBlockSize(4);
-		std::vector< std::vector<double> > xy;
-		std::vector<double> atts(nl);
-		for (size_t i=0; i<out.bs.n; i++) {
-			std::vector<double> vals = readBlock(out.bs, i);
-			unsigned nc=out.bs.nrows[i] * ncol();
-			for (size_t j=0; j<nc; j++) {
-				for (size_t k=0; k<nl; k++) {
-					size_t kk = j + k * nl;
-					att[nc+j][k] = vals[kk];
-				}
-				xy = xyFromCell(nc+j);
-				getCorners(x, y, xy[0][0], xy[1][0], xr, yr);
-				SpatPart p(x, y);
-				g.addPart(p);
-				v.addGeom(g);
-				g.parts.resize(0);
-
-			}
-		}
-		SpatDataFrame df;
-		std::vector<std::string> nms = getNames();
-		for (size_t i=0; i<att.size(); i++) {
-			df.add_column(att[i], nms[i]);
-		}
-	}
-	v.setCRS(getCRS());
-	return(v);
-}
-
-*/
-
-SpatVector SpatRaster::as_polygons(bool trunc, bool dissolve, bool values, bool narm, bool nall, SpatOptions &opt) {
+SpatVector SpatRaster::as_polygons(bool round, bool dissolve, bool values, bool narm, bool nall, int digits, SpatOptions &opt) {
 
 	if (!hasValues()) {
 		values = false;
@@ -2246,7 +2349,7 @@ SpatVector SpatRaster::as_polygons(bool trunc, bool dissolve, bool values, bool
 	}
 
 	if (dissolve) {
-		return polygonize(trunc, values, narm, dissolve, opt);
+		return polygonize(round, values, narm, dissolve, digits, opt);
 	}
 
 	SpatVector vect;
@@ -2324,9 +2427,9 @@ SpatVector SpatRaster::as_polygons(bool trunc, bool dissolve, bool values, bool
 
 	std::reverse(std::begin(vect.geoms), std::end(vect.geoms));
 
-	if (dissolve) {
-		vect = vect.aggregate(vect.get_names()[0], true);
-	}
+//	if (dissolve) {
+//		vect = vect.aggregate(vect.get_names()[0], true);
+//	}
 
 	if (remove_values) {
 		vect.df = SpatDataFrame();
@@ -2404,7 +2507,13 @@ bool SpatRaster::setRGB(int r, int g, int b, int alpha, std::string type) {
 		size_t mnlyr =  vmin( channels, false );;
 		if (mnlyr >= 0) {
 			rgblyrs = channels;
-			rgbtype = type; // check validity
+			std::vector<std::string> f = {"rgb", "hsv", "hsi", "hsl"};
+			std::transform(type.begin(), type.end(), type.begin(), ::tolower);
+			if (std::find(f.begin(), f.end(), type) == f.end()) {
+				addWarning("color type must be one of: 'rgb', 'hsv', 'hsi', 'hsl'");
+				type = "rgb";
+			}
+			rgbtype = type; 
 			rgb = true;
 		} else {
 			rgb = false;
@@ -2420,6 +2529,7 @@ std::vector<int> SpatRaster::getRGB(){
 
 void SpatRaster::removeRGB(){
 	rgblyrs = std::vector<int>(0);
+	rgbtype = "";
 	rgb = false;
 }
 
@@ -2461,3 +2571,101 @@ std::vector<int> SpatRaster::getFileBlocksize() {
 }
 
 
+bool SpatRaster::addTag(std::string name, std::string value) {
+	lrtrim(name);
+	lrtrim(value);
+	if (value == "") {
+		return removeTag(name);
+	} else if (name != "") {
+		tags[name] = value;
+		return true;
+	} 
+	return false;
+}
+
+bool SpatRaster::removeTag(std::string name) {
+	std::map<std::string, std::string>::iterator it = tags.find(name);
+	if (it == tags.end()) return false;
+	tags.erase(it);
+	return true;
+}
+
+std::string SpatRaster::getTag(std::string name) {
+	std::map<std::string, std::string>::iterator it = tags.find(name);
+	if (it != tags.end()) return it->second;
+	return "";
+}
+
+std::vector<std::string> SpatRaster::getTags() {
+	std::vector<std::string> out;
+	out.reserve(2 * tags.size());
+	for(auto e : tags) {
+		out.push_back(e.first);
+		out.push_back(e.second);
+	}
+	return out;
+}
+
+
+
+void SpatRaster::addLyrTags(std::vector<size_t> lyrs, std::vector<std::string> names, std::vector<std::string> values) {
+
+	size_t n = std::max(std::max(lyrs.size(), names.size()), values.size());
+	if (n == 0) return;
+	
+	recycle(lyrs, n);
+	recycle(names, n);
+	recycle(values, n);
+	
+	size_t nl = nlyr();
+	for (size_t i=0; i<lyrs.size(); i++) {
+		if (lyrs[i] >= nl) continue;
+		lrtrim(names[i]);
+		lrtrim(values[i]);
+		if (values[i] == "") {
+			removeLyrTag(lyrs[i], names[i]);
+		} else {
+			if (lyrs[i] >= lyrTags.size()) lyrTags.resize(lyrs[i]+1);
+			if (names[i] != "") {
+				lyrTags[lyrs[i]][names[i]] = values[i];
+			} 
+		}
+	}
+}
+
+bool SpatRaster::removeLyrTag(size_t lyr, std::string name) {
+	if (lyr >= lyrTags.size()) return false;
+	std::map<std::string, std::string>::iterator it = lyrTags[lyr].find(name);
+	if (it == lyrTags[lyr].end()) return false;
+	lyrTags[lyr].erase(it);
+	return true;
+}
+
+bool SpatRaster::removeLyrTags() {
+	lyrTags.resize(0);
+	return true;
+}
+
+
+std::string SpatRaster::getLyrTag(size_t lyr, std::string name) {
+	if (lyr >= lyrTags.size()) return "";
+	std::map<std::string, std::string>::iterator it = lyrTags[lyr].find(name);
+	if (it != lyrTags[lyr].end()) return it->second;
+	return "";
+}
+
+std::vector<std::string> SpatRaster::getLyrTags(std::vector<size_t> lyrs) {
+	std::vector<std::string> out;
+	out.reserve(lyrs.size());
+	for (size_t i=0; i<lyrs.size(); i++) {
+		if (lyrs[i] < lyrTags.size()) {
+			for(auto e : lyrTags[lyrs[i]]) {
+				out.push_back(std::to_string(lyrs[i]));
+				out.push_back(e.first);
+				out.push_back(e.second);
+			}
+		}
+	}
+	return out;
+}
+
