@@ -43,7 +43,7 @@ void add_quotes(std::vector<std::string> &s) {
 
 std::string quoted_csv(const std::vector<std::string> &s) {
 	std::string ss;
-	if (s.size() == 0) {
+	if (s.empty()) {
 		ss = "";
 		return ss;
 	}
@@ -160,6 +160,27 @@ bool is_rat(SpatDataFrame &d) {
 	return true;
 }
 
+
+bool is_ratct(SpatDataFrame &d) {
+
+	std::vector<std::string> ss = {"red", "green", "blue", "r", "g", "b"};
+	std::vector<std::string> nms = d.names;
+	size_t cnt = 0;
+	
+	for (size_t i=0; i<nms.size(); i++) {
+		std::string name = nms[i];
+		lowercase(name);
+		int k = where_in_vector(name, ss, true);
+		if (k >= 0) {
+			cnt++;
+		}
+	}
+	return (cnt >= 3);
+}
+
+
+
+
 /*
 bool setCats(GDALRasterBand *poBand, std::vector<std::string> &labels) {
 	char **labs = NULL;
@@ -310,17 +331,17 @@ void removeVatJson(std::string filename) {
 bool SpatRaster::writeStartGDAL(SpatOptions &opt, const std::vector<std::string> &srcnames) {
 
 	std::string filename = opt.get_filename();
-	if (filename == "") {
+	
+	if (filename.empty()) {
 		setError("empty filename");
 		return(false);
-	} else {
-		// make sure filename won't be used again
-		opt.set_filenames({""});
-	}
+	} 
+	// assure filename won't be used again
+	opt.set_filenames({""});
 
 	std::string driver = opt.get_filetype();
 	getGDALdriver(filename, driver);
-	if (driver == "") {
+	if (driver.empty()) {
 		setError("cannot guess file type from filename");
 		return(false);
 	}
@@ -340,9 +361,9 @@ bool SpatRaster::writeStartGDAL(SpatOptions &opt, const std::vector<std::string>
 	std::string datatype = opt.get_datatype();
 
 	bool writeRGB = (rgb && nlyr() == 3 && rgblyrs.size() == 3);
-	if (writeRGB) {
-		datatype = "INT1U";
-	}
+	//if (writeRGB) {
+	//	datatype = "INT1U";
+	//}
 	std::string errmsg;
 	if (!checkFormatRequirements(driver, filename, errmsg)) {
 		setError(errmsg);
@@ -386,7 +407,9 @@ bool SpatRaster::writeStartGDAL(SpatOptions &opt, const std::vector<std::string>
 		// other layers affected? etc.
 		warnCT = false;
 		if (hasCT[0]) {
-			if (ct[0].nrow() < 256) {
+			if (is_ratct(source[0].cats[0].d)) {
+				std::fill(hasCT.begin(), hasCT.end(), false);
+			} else if (ct[0].nrow() < 256) {
 				if (opt.datatype_set && (datatype != opt.get_datatype())) {
 					addWarning("change datatype to INT1U to write the color-table");
 				} else {
@@ -406,7 +429,7 @@ bool SpatRaster::writeStartGDAL(SpatOptions &opt, const std::vector<std::string>
 				datatype = "INT4S";
 			}
 		}
-	} else if (hasCT[0] || cat) {
+	} else if (hasCT[0]) {
 		if (opt.datatype_set && (datatype != "INT1U")) {
 			addWarning("change datatype to INT1U to write the color-table");
 		} else {
@@ -500,11 +523,11 @@ bool SpatRaster::writeStartGDAL(SpatOptions &opt, const std::vector<std::string>
 
 	CSLDestroy( papszOptions );
 	if (poDS == NULL) {
-		if (!filepath_exists(filename)) {
-			setError("failed writing "+ driver + " file. Path does not exist:\n   " + filename);
-		} else {
+		//if (!filepath_exists(filename)) {
+		//	setError("failed writing "+ driver + " file. Path does not exist:\n   " + filename);
+		//} else {
 			setError("failed writing "+ driver + " file");
-		}
+		//}
 		GDALClose( (GDALDatasetH) poDS );
 		return false;
 	}
@@ -575,6 +598,15 @@ bool SpatRaster::writeStartGDAL(SpatOptions &opt, const std::vector<std::string>
 	}
 
 	bool scoffwarning = false;
+
+	if (driver == "GTiff") {
+		std::vector<std::string> m = getTags();
+		if (m.size() > 0) {
+			for (size_t i=0; i<m.size(); i+=2) {
+				poDS->SetMetadataItem(m[i].c_str(), m[i+1].c_str(), "USER_TAGS");
+			}
+		}
+	}
 	
 	for (size_t i=0; i < nlyr(); i++) {
 
@@ -613,6 +645,16 @@ bool SpatRaster::writeStartGDAL(SpatOptions &opt, const std::vector<std::string>
 		*/
 		poBand->SetDescription(nms[i].c_str());
 
+		if (driver == "GTiff") {
+			std::vector<std::string> m = getLyrTags({i});
+			if (m.size() > 0) {
+				for (size_t i=0; i<m.size(); i+=2) {
+					poBand->SetMetadataItem(m[i].c_str(), m[i+1].c_str());
+				}
+			}
+		}
+
+
 		if ((i==0) || (driver != "GTiff")) {
 			// to avoid "Setting nodata to nan on band 2, but band 1 has nodata at nan."
 			if (hasNAflag) {
@@ -975,12 +1017,14 @@ bool SpatRaster::writeStopGDAL() {
 		}
 	}
 
-	if (copy_driver != "") {
+	if (copy_driver.empty()) {
+		GDALClose( (GDALDatasetH) source[0].gdalconnection );
+	} else {
 		GDALDataset *newDS;
 		GDALDriver *poDriver;
 		char **papszOptions = set_GDAL_options(copy_driver, 0.0, false, gdal_options);
 		poDriver = GetGDALDriverManager()->GetDriverByName(copy_driver.c_str());
-		if (copy_filename == "") {
+		if (copy_filename.empty()) {
 			newDS = poDriver->CreateCopy(source[0].filename.c_str(),
 				source[0].gdalconnection, FALSE, papszOptions, NULL, NULL);
 			if( newDS == NULL )  {
@@ -1022,8 +1066,6 @@ bool SpatRaster::writeStopGDAL() {
 			GDALClose( (GDALDatasetH) newDS );
 		}
 		CSLDestroy(papszOptions);
-	} else {
-		GDALClose( (GDALDatasetH) source[0].gdalconnection );
 	}
 	source[0].hasValues = true;
 
